from dataclasses import dataclass, field
import numpy as np
from typing import List, Union

# Internal Imports
from brom_drake.productions.types.base.configuration import Configuration as BaseConfiguration
from .script import Script as AttemptGraspWithPuppeteerWristScript

@dataclass
class Configuration:
    """
    *Description*

    A dataclass that specifies the configuration for the
    "attempt_grasp_with_puppeteer_wrist" production.

    *Attributes*

    base: BaseConfiguration, optional
        The base configuration for the production.

    show_collision_geometries: bool, optional
        A boolean flag that indicates whether or not to show collision geometries.
        Default is False.
        (TODO(Kwesi): Implement this feature)

    initial_gripper_joint_positions: List[float] or np.ndarray
        A list or array of floats representing the initial joint positions of the gripper.
        Default is None.

    target_body_on_gripper: str, optional
        The name of the body on the gripper that we use to define X_GripperObject.
        Default is None.
        (TODO(Kwesi): Implement this feature)

    gripper_color: List[float], optional
        A list of 4 floats representing the RGBA color of the gripper.
        Default is None.
        (TODO(Kwesi): Implement this feature)

    show_gripper_base_frame: bool, optional
        A boolean flag that indicates whether or not to show the gripper base frame.
        Default is False.
        (TODO(Kwesi): Consider deleting this feature)

    script: AttemptGraspWithPuppeteerWristScript, optional
        The script configuration for the production.
        Default is the standard AttemptGraspWithPuppeteerWristScript.

    """
    base: BaseConfiguration = field(default_factory=BaseConfiguration) #By default, use the standard script
    show_collision_geometries: bool = False
    initial_gripper_joint_positions: Union[List[float], np.ndarray] = None
    target_body_on_gripper: str = None # The "Gripper" Frame that we use to define X_GripperObject
    gripper_color: List[float] = None
    show_gripper_base_frame: bool = False
    script: AttemptGraspWithPuppeteerWristScript = field(default_factory=AttemptGraspWithPuppeteerWristScript) #By default, use the standard script

    @property
    def meshcat_port_number(self):
        """
        *Description*

        A property that gets the meshcat port number from the base configuration.
        """
        return self.base.meshcat_port_number

    @meshcat_port_number.setter
    def meshcat_port_number(self, new_value: int):
        """
        *Description*

        A property setter that sets the meshcat port number in the base configuration.
        """
        self.base.meshcat_port_number = new_value


import numpy as np
from pydrake.all import (
    AbstractValue, BasicVector,
    Context,
    LeafSystem,
    RigidTransform,
)

from .configuration import Configuration as RigidTransformToVectorSystemConfiguration

class RigidTransformToVectorSystem(LeafSystem):
    """
    *Description*
    
    This system will take in a RigidTransform and output a vector
    of size 7, where the first 3 elements are the translation
    and the last 4 elements are the quaternion representation
    of the rotation.

    *Diagram*

    The LeafSystem's inputs and outputs can be represented as: ::
    
                                |-------------------|
                                | RigidTransform    |
        rigid_transform ---->   | To                | vector_xyz_quat(wxyz) (BasicVector[7])
        (RigidTransform)        | VectorSystem      |
                                |-------------------|
    """
    def __init__(
        self,
        config: RigidTransformToVectorSystemConfiguration = None,
    ):
        # Input Processing
        if config is None:
            config = RigidTransformToVectorSystemConfiguration()
        self.config = config

        # Base class initialization
        LeafSystem.__init__(self)
        self.set_name(self.config.name)

        # Create inputs
        self.create_system_inputs()

    def CalcVectorOutput(self, context: Context, output: BasicVector):
        # Setup
        config: RigidTransformToVectorSystemConfiguration = self.config

        # Collect Transform
        rigid_transform: RigidTransform = self.rigid_transform_input.Eval(context)
        
        # Create output vector
        if config.output_format == "vector_xyz_quat(wxyz)":
            xyz = rigid_transform.translation()
            quat = rigid_transform.rotation().ToQuaternion().wxyz()
            output.SetFromVector(
                np.hstack((xyz, quat))
            )
        elif config.output_format == "vector_xyz_euler(rpy)":
            xyz = rigid_transform.translation()
            euler = rigid_transform.rotation().ToRollPitchYaw().vector()
            output.SetFromVector(
                np.hstack((xyz, euler))
            )

    def create_system_inputs(self):
        # Setup
        config: RigidTransformToVectorSystemConfiguration = self.config

        # Create ports
        self.rigid_transform_input = self.DeclareAbstractInputPort(
            "rigid_transform",
            AbstractValue.Make(RigidTransform()),
        )

        if config.output_format == "vector_xyz_quat(wxyz)":
            self.DeclareVectorOutputPort(
                "vector_xyz_quat(wxyz)",
                7,
                self.CalcVectorOutput,
            )
        elif config.output_format == "vector_xyz_euler(rpy)":
            self.DeclareVectorOutputPort(
                "vector_xyz_euler(rpy)",
                6,
                self.CalcVectorOutput,
            )
        else:
            raise ValueError(f"Unknown output format: {config.output_format}")



from pydrake.all import (
    AbstractValue,
    BasicVector,
    Context,
    InputPort,
    InputPortIndex,
    LeafSystem,
)
from typing import Union

class FlexiblePortSwitch(LeafSystem):
    """
    *Description*
    
    This system is a simplification of the PortSwitch system primitive
    that is built into Drake. It allows for you to command the switch
    with a simple input (i.e., a string or an int) instead of
    a more complicated type.
    """

    def __init__(self, dim: int, selector_type_in: type = str):
        LeafSystem.__init__(self)
        
        # Setup
        self.dim = dim
        self.selector_type_in = selector_type_in

        # Initialize the system with no input ports
        self._input_ports = {}
        self._current_input_port = None

        # Declare the port selector input port
        self.declare_port_selector_input_port("port_selector", selector_type_in)

        # Declare the value output port
        self.DeclareVectorOutputPort(
            "value",
            self.dim,
            self.DoCalcValue,
        )

    def declare_port_selector_input_port(self, name: str, type_in: type = str) -> InputPort:
        # Setup

        # Conditional based on type_in
        if type_in == str:
            self.port_selector_port = self.DeclareAbstractInputPort(
                "port_selector",
                AbstractValue.Make("model_name"),
            )
        elif type_in == int:
            self.port_selector_port = self.DeclareVectorInputPort(
                name="port_selector",
                size=1,
            )
        elif type_in == InputPortIndex:
            self.port_selector_port = self.DeclareAbstractInputPort(
                "port_selector",
                AbstractValue.Make(InputPortIndex),
            )
        else:
            raise NotImplementedError(
                f"Type {type_in} is not supported. Please use str."
            )

    def DeclareInputPort(self, name: str) -> InputPort:
        """
        Description
        -----------
        This method will declare an input port for the system.
        """
        # Setup
        selector_type_in = self.selector_type_in

        # Create the input port
        input_port = self.DeclareVectorInputPort(
            name=name,
            size=self.dim,
        )

        # Update the input ports dictionary
        if selector_type_in == str:
            self._input_ports[name] = input_port
        elif selector_type_in == int:
            self._input_ports[len(self._input_ports.keys())] = input_port
        elif selector_type_in == InputPortIndex:
            self._input_ports[input_port.get_index()] = input_port
        else:
            raise NotImplementedError(
                f"Type {selector_type_in} is not supported. Please use str, int, or InputPortIndex."
            )
        
        return input_port
    
    def get_port_selector_input_port(self) -> InputPort:
        """
        Description
        -----------
        This method will return the port selector input port.
        """
        return self.port_selector_port
    
    def DoCalcValue(self, context: Context, output: BasicVector):
        """
        Description
        -----------
        This method will calculate the value of the output port.
        """
        # Setup
        selector_type_in = self.selector_type_in

        # Get the port selector value
        port_selection = self.port_selector_port.Eval(context)
        if selector_type_in == int:
            port_selection = int(port_selection[0])

        # Vet the port selector value
        assert isinstance(port_selection, self.selector_type_in), \
            f"Expected {self.selector_type_in}; received {type(port_selection)}"
        assert len(self._input_ports) > 0, \
            "No input ports have been declared. Please declare at least one input port."

        # Get the input port
        input_port = self._input_ports[port_selection]

        # Get the value from the input port
        value = input_port.Eval(context)

        # Set the output value
        output.SetFromVector(value)


from .base import BaseRRTPlannerConfig, BaseRRTPlanner
from .bidirectional import BidirectionalRRTPlannerConfig, BidirectionalRRTPlanner
from .bidirectional_connect import BidirectionalRRTConnectPlannerConfig, BidirectionalRRTConnectPlanner
from .connect import RRTConnectPlannerConfig, RRTConnectPlanner

__all__ = [
    "BaseRRTPlannerConfig",
    "BaseRRTPlanner",
    "BidirectionalRRTPlannerConfig",
    "BidirectionalRRTPlanner",
    "BidirectionalRRTConnectPlannerConfig",
    "BidirectionalRRTConnectPlanner",
    "RRTConnectPlannerConfig",
    "RRTConnectPlanner",
]


from importlib import resources as impresources
from typing import List, Tuple, Callable

import networkx as nx
import numpy as np
from pydrake.common.eigen_geometry import Quaternion
from pydrake.common.value import AbstractValue
from pydrake.geometry import GeometrySet, CollisionFilterDeclaration
from pydrake.math import RigidTransform, RollPitchYaw
from pydrake.multibody.parsing import Parser
from pydrake.multibody.plant import MultibodyPlant
from pydrake.systems.framework import Diagram, Context
from pydrake.systems.primitives import ConstantValueSource

# Internal Imports
import brom_drake.robots as robots
from brom_drake.motion_planning.systems.open_loop_dispensers.open_loop_plan_dispenser import OpenLoopPlanDispenser
from brom_drake.stations.kinematic import UR10eStation as KinematicUR10eStation
from brom_drake.productions import ProductionID
from brom_drake.productions.roles import Role
from brom_drake.productions.types.motion_planning import KinematicMotionPlanningProduction
from brom_drake.utils import Performer, GroundShape, AddGround, MotionPlan


class ShelfPlanning1(KinematicMotionPlanningProduction):
    """
    *Description*

    This production contains a simple motion planning task
    where we attempt to reach to different places around the shelfs
    of a bookcase.

    This is a Kinematic motion planning production, which means that the
    the robot's actuators will not be driven. Instead, the robot's links will
    be reset to "exactly" the configurations in the motion plan. This is not realistic
    but tests the correctness of the plan without worrying about the control problem
    of moving the arm.

    *Parameters*

    time_step: float, optional
        The time step for the production, by default 1e-3.

    shelf_pose: RigidTransform, optional
        The pose of the shelf in the world frame, by default None.

    meshcat_port_number: int, optional
        The port number for meshcat visualization, by default 7001.

    plan_execution_speed: float, optional
        The speed at which to execute the motion plan, by default 0.2

    *Usage*

    You may use this in a similar way to all other KinematicMotionPlanningProduction
    objects. ::

        # Other imports above...
        from brom_drake.motion_planning.algorithms.rrt import RRTConnectPlannerConfig, RRTConnectPlanner
        from brom_drake.productions.motion_planning.offline import ShelfPlanning1

        # Define the goal pose
        easy_goal_position = np.array([+0.0, 1.0, 1.05])
        goal_orientation = RollPitchYaw(np.pi / 2.0, np.pi / 2.0, 0.0).ToQuaternion()
        goal_pose = RigidTransform(goal_orientation, easy_goal_position)

        # Create the production
        production = ShelfPlanning1(
            meshcat_port_number=meshcat_port_number, # Use None for CI
            goal_pose=goal_pose,
        )

        # Create a planner object which will be used to plan the motion
        config = RRTConnectPlannerConfig(
            steering_step_size=0.01,
            prob_sample_goal=0.05,
            max_iterations=int(1e5),
            convergence_threshold=1e-3,
        )
        planner2 = RRTConnectPlanner(
            production.arm,
            production.plant,
            production.scene_graph,
            config=config,
        )

        # To build the production, we only need to provide a planning function
        # (can come from anywhere, not just a BaseRRTPlanner object)
        diagram, diagram_context = production.easy_cast_and_build(
            planner2.plan,
            with_watcher=True,
        )

        print("Simulating...")

        # Simulate the diagram
        simulator = Simulator(diagram, diagram_context)
        simulator.set_target_realtime_rate(1.0)

        # Run simulation
        simulator.Initialize()
        simulator.AdvanceTo(0.1)
        planned_trajectory = production.plan_dispenser.planned_trajectory
        print(f"Expected end time of the planned trajectory: {planned_trajectory.end_time()}")
        # return
        simulator.AdvanceTo(planned_trajectory.end_time()+1.0)


    """
    def __init__(
        self,
        time_step: float = 1e-3,
        shelf_pose: RigidTransform = None, # The pose of the shelf
        meshcat_port_number: int = 7001, # Usually turn off for CI (i.e., make it None)
        plan_execution_speed: float = 0.2,
        **kwargs,
    ):
        """
        *Description*

        Constructor for ShelfPlanning1.
        
        This:
        
        - Stores all inputs in the appropriate places
        - Selects the default shelf_pose, if none is given
        - Sets the cupboard joints to be in the "open" configuration
        - constructs the KinematicUR10eStation used in testing
        - Sets the names of the plant and scene graph to use the ShelfPlanning tag

        *Parameters*

        time_step: float, optional
            The time step for the production, by default 1e-3.

        shelf_pose: RigidTransform, optional
            The pose of the shelf in the world frame, by default None.

        meshcat_port_number: int, optional
            The port number for meshcat visualization, by default 7001.

        plan_execution_speed: float, optional
            The speed at which to execute the motion plan, by default 0.2        
        """
        super().__init__(**kwargs)

        self.time_step = time_step
        self.shelf_pose = shelf_pose
        self.meshcat_port_number = meshcat_port_number
        self.plan_execution_speed = plan_execution_speed

        # Input Processing
        if self.shelf_pose is None:
            self.shelf_pose = RigidTransform(
                RollPitchYaw(0.0, 0.0, +np.pi/2.0).ToQuaternion(),
                np.array([0.0, 1.0, 0.65]),
            )

        self.desired_cupboard_positions = np.array([0.2, 0.3])

        # Create containers for meshcat and other values we will set later
        self.meshcat = None
        self.plan_dispenser = None

        # Set station
        self.station = KinematicUR10eStation(
            time_step=self.time_step,
            meshcat_port_number=self.meshcat_port_number,
        )
        self.arm = self.station.arm
        self.robot_model_idx_ = self.station.arm

        # Set Names of Plant and scene graph
        self.plant = self.station.plant
        self.scene_graph = self.station.scene_graph

        self.plant.set_name(f"Shelf1_Production_Plant")
        self.scene_graph.set_name(f"Shelf1_Production_SceneGraph")

        # Create variables for the models we have
        self.shelf_model_index = None
        self.geometry_ids_to_ignore = []

    def add_supporting_cast(self):
        """
        *Description*

        Modifies the plant and the builder by:
        - Adding UR10e station to the plant
        - Adding shelf to the plant
        - Adding a "ground" to the plant
        - Adding features required by the KinematicMotionPlanningProductionFeatures
          + Robot's ModelInstanceIndex Source
          + Motion Planning Components
          + Start and Goal Pose sources
        - Setting initial configuration of the arm for the plant
        
        """
        # Call the parent class method
        super().add_supporting_cast()

        # Setup
        self.add_ur10e_station() # Use the UR10e station's plant + scene graph for all other objects

        # Add obstacles
        self.add_shelf(self.plant)

        # Add the ground as well as the start and goal locations
        AddGround(self.plant)
        self.station.Finalize()

        # Add The Motion Planning Components (e.g., the interpolator)
        self.add_robot_source_system()

        self.add_motion_planning_components()

        self.add_start_and_goal_sources_to_builder()

        # Add initial conditions for the arm
        self.initial_condition_manager.add_initial_configuration(
            model_instance_index=self.arm,
            configuration=self.start_configuration,
        )
        # The initial condition will be set for the plant when we build the production.

        # Print message to user
        print("Added all secondary cast members to the builder.")

    def add_motion_planning_components(self):
        """
        *Description*

        Connects the plan dispenser's output port (i.e., the reference signal for the
        arm's joint controller) to the input of the UR10eStation.

        """
        # Setup
        n_actuated_dof = self.plant.num_actuated_dofs()

        # Add the Plan Dispenser and connect it to the station
        self.plan_dispenser = self.builder.AddSystem(
            OpenLoopPlanDispenser(n_actuated_dof, self.plan_execution_speed)
        )

        self.builder.Connect(
            self.plan_dispenser.GetOutputPort("point_in_plan"),
            self.station.GetInputPort("desired_joint_positions"),
        )

    def add_shelf(self, plant: MultibodyPlant):
        """
        *Description*

        Adds a model of the cupboard to the production's plant.

        *Parameter*

        plant: pydrake.multibody.plant.MultibodyPlant
            The plant that we want to add the cupboard to.
        """
        # Setup
        urdf_file_path = str(
            impresources.files(robots) / "models/cupboard/cupboard.sdf"
        )

        # Add the shelf to the plant
        model_idcs = Parser(plant=plant).AddModels(urdf_file_path)
        self.shelf_model_index = model_idcs[0]

        # Weld the bookshelf to the world frame
        plant.WeldFrames(
            plant.world_frame(),
            plant.GetFrameByName("cupboard_body", self.shelf_model_index),
            self.shelf_pose,
        )

    def add_ur10e_station(self):
        """
        *Description*
        
        Add the Kinematic UR10e station to the production's plant for simple motion planning.
        """
        # Setup

        # Add station
        self.builder.AddSystem(self.station)

    def add_cast_and_build(
        self,
        cast: List[Tuple[Role, Performer]] = [],
        with_watcher: bool = False,
    ) -> Tuple[Diagram, Context]:
        """
        *Description*
        
        Modifies the normal add_cast_and_build, so that
        we share the context of the plant with the appropriate
        parts of the system.

        In addition to using the parent class's implementation of add_cast_and_build, this:
        - Configures the collision filter to avoid spurious collisions between the cupboard's 
        various parts and itself
        - Insert the motion planner role into the diagram and connect it properly

        TODO(kwesi): Consider making this a standard part of the
        OfflineKinematicMotionPlanningProduction class.

        *Parameters*

        cast: List[Tuple[Role, Performer]], optional
            The cast to add to the production, by default [].

        with_watcher: bool, optional
            A Boolean that determines whether to add a watcher to the diagram, by default False.

        *Returns*

        diagram: Diagram
            The built diagram.

        diagram_context: Context
            The context for the built diagram.
            This context will be used to initialize the watcher.
        """
        diagram, diagram_context = super().add_cast_and_build(
            main_cast_members=cast,
            with_watcher=with_watcher,
        )

        # Configure the scene graph for collision detection
        self.configure_collision_filter(
            diagram.GetSubsystemContext(
                self.scene_graph, diagram_context,
            )
        )

        # Connect arm controller to the appropriate plant_context
        self.station.arm_controller.plant_context = diagram.GetSubsystemContext(
            self.station.arm_controller.plant, diagram_context,
        )

        for role_ii, performer_ii in cast:
            if role_ii.name == "OfflineMotionPlanner":
                performer_ii.set_internal_root_context(
                    diagram_context
                )

        return diagram, diagram_context

    def configure_collision_filter(self, scene_graph_context: Context):
        """
        *Description*
        
        This method configures the collision filter, so that:
        
        - self collisions between the shelf's pieces, and
        - self collisions between the arm's parts,
        
        are ignored during simulation.

        *Parameters*

        scene_graph_context: pydrake.systems.framework.Context
            The context of the scenegraph for the constructed production (i.e., the diagram
            for the built Production).
        """
        # Setup
        scene_graph = self.scene_graph
        shelf_model_index = self.shelf_model_index

        # Ignore self collisions of the shelf, by:
        # - Getting the model instance index of the shelf
        # - Collecting the Geometry IDs of all the geometries in the shelf
        shelf_geometry_ids = []
        for body_index in self.plant.GetBodyIndices(shelf_model_index):
            # Get the geometry IDs
            shelf_geometry_ids.extend(
                self.plant.GetCollisionGeometriesForBody(
                    self.plant.get_body(body_index)
                )
            )
            # print(self.plant.get_body(body_index))

        self.shelf_geometry_ids = shelf_geometry_ids

        # Apply the collision filter
        scene_graph.collision_filter_manager(scene_graph_context).Apply(
            CollisionFilterDeclaration().ExcludeWithin(
                GeometrySet(self.shelf_geometry_ids)
            )
        )

        # Ignore collisions between the robot links
        # TODO: Actually implement this for adjacent links? Or is this not necessary?
        arm_geometry_ids = []
        for body_index in self.plant.GetBodyIndices(self.arm):
            arm_geometry_ids.extend(
                self.plant.GetCollisionGeometriesForBody(
                    self.plant.get_body(body_index)
                )
            )

        self.arm_geometry_ids = arm_geometry_ids

        # Apply the collision filter
        scene_graph.collision_filter_manager(scene_graph_context).Apply(
            CollisionFilterDeclaration().ExcludeWithin(
                GeometrySet(self.arm_geometry_ids)
            )
        )

    def easy_cast_and_build(
        self,
        planning_algorithm: Callable[
            [np.ndarray, np.ndarray, Callable[[np.ndarray], bool]],
            Tuple[MotionPlan, bool],
        ],
        with_watcher: bool = False,
    ) -> Tuple[Diagram, Context]:
        """
        *Description*
        
        This function is used to easily cast and build the production.

        *Parameters*

        planning_algorithm: Callable[[np.ndarray, np.ndarray, Callable[[np.ndarray], bool]], Tuple[MotionPlan, bool]
            The algorithm that produces a path from the start configuration to goal configuration
            using the collision checker for the scene.

        with_watcher: bool, optional
            A Boolean that determines whether to add a watcher to the diagram.
            Default is False.
        
        *Returns*

        diagram: pydrake.systems.framework.Diagram
            The diagram that represents the fully assembled cast, all connected together.

        diagram_context: pdrake.systems.framework.Context
            The context for the diagram (used by the diagram watcher to monitor the system, if defined).

        .. note::

            This method returns a `pydrake.systems.framework.Diagram` and its associated `pydrake.systems.framework.Context`.
            When constructing a Drake simulation of this, you should use these objects to create the `pydrake.systems.analysis.Simulator`.
        """
        # Setup

        # Use Base class implementation to start
        diagram, diagram_context = super().easy_cast_and_build(
            planning_algorithm,
            with_watcher=with_watcher,
        )

        # Configure the scene graph for collision detection
        self.configure_collision_filter(
            diagram.GetSubsystemContext(
                self.scene_graph, diagram_context,
            )
        )

        # Connect arm controller to the appropriate plant_context
        self.station.arm_controller.plant_context = diagram.GetSubsystemContext(
            self.station.arm_controller.plant, diagram_context,
        )

        self.performers[0].set_internal_root_context(
            diagram_context
        )

        # Set the initial positions of the arm
        # self.station.plant.SetPositions(
        #     self.station.plant.GetMyMutableContextFromRoot(diagram_context),
        #     self.arm,
        #     self.start_configuration,
        # )

        return diagram, diagram_context

    @property
    def goal_pose(self) -> RigidTransform:
        """
        *Description*

        Get the goal pose. 
        
        *Returns*
        
        pose_WorldGoal: RigidTransform
            The goal pose.
        """

        # Algorithm
        if self._goal_pose is not None:
            # If we have already defined the goal pose, then return it
            return self._goal_pose
        elif (self._goal_pose is None) and (self._goal_config is None):
            # If we have no guidance on how to start, then we will use a default
            goal_position = np.array([+0.0, 1.0, 0.6])
            goal_orientation = RollPitchYaw(np.pi / 2.0, np.pi / 2.0, 0.0).ToQuaternion()
            self._goal_pose = RigidTransform(goal_orientation, goal_position)
            return self._goal_pose
        elif (self._goal_pose is None) and (self._goal_config is not None):
            # Use the goal configuration to get the goal pose
            # Using a "forward kinematics solver"
            self._goal_pose = self.solve_forward_kinematics_problem_for_arm(self._goal_config)
            return self._goal_pose
        else:
            raise ValueError(
                f"Unexpected behavior. This should never happen."
            )


    @property
    def id(self) -> ProductionID:
        """Always returns ``ProductionID.kShelfPlanning1``"""
        return ProductionID.kShelfPlanning1

    def solve_forward_kinematics_problem_for_arm(
        self,
        robot_joint_positions: np.ndarray,
        target_frame_name: str = "ft_frame",
    ) -> RigidTransform:
        """
        *Description*
        
        This method solves the forward kinematics problem for the UR10e arm
        by itself.

        TODO(Kwesi): Attempt to make this a utility (it is duplicated multiple times!)

        *Parameters*

        robot_joint_positions: np.ndarray of shape (6,)
            The joint configuration of the robot in the production.

        target_frame_name: str, optional
            The frame that we compute the pose of with respect to the world origin frame.
            Default is "ft_frame"
        
        *Returns*
        
        pose_WorldTarget: RigidTransform
            The end effector pose of the robot expressed in the world frame.
        """
        # Setup

        # Create shadow plant and populate it with the UR10e arm
        shadow_station = KinematicUR10eStation(
            time_step=self.time_step,
            meshcat_port_number=None,
        )
        shadow_station.Finalize()

        # Use station's plant to solve the forward kinematics problem
        shadow_plant = shadow_station.plant
        
        temp_context = shadow_plant.CreateDefaultContext()
        shadow_plant.SetPositions(
            temp_context,
            shadow_station.arm,
            robot_joint_positions,
        )

        # Get the pose of the end effector
        end_effector_pose = shadow_plant.EvalBodyPoseInWorld(
            temp_context,
            shadow_plant.GetBodyByName(target_frame_name),
        )

        return end_effector_pose

    @property
    def start_pose(self) -> RigidTransform:
        """
        *Description*
        
        Get the start pose.
        
        *Returns*

        pose_WorldStart: RigidTransform
            The start pose of the robot's end effector frame with respect to the world frame.
        """
        # Setup

        if self._start_pose is not None:
            # If we have already defined the start pose, then return it
            return self._start_pose
        elif (self._start_pose is None) and (self._start_config is None):
            # If we have no guidance on how to start, then we will use a default
            start_position = np.array([+0.3, 0.1, 1.2])
            start_orientation = Quaternion(1, 0, 0, 0)
            self._start_pose = RigidTransform(start_orientation, start_position)
            return self._start_pose
        elif (self._start_pose is None) and (self._start_config is not None):
            # Use the start configuration to get the starting pose
            # Using a "forward kinematics solver"
            self._start_pose = self.solve_forward_kinematics_problem_for_arm(self._start_config)
            return self._start_pose
        else:
            raise ValueError(
                f"Unexpected behavior. This should never happen."
            )


import numpy as np
from pydrake.multibody.plant import MultibodyPlant
from pydrake.multibody.tree import ModelInstanceIndex
from pydrake.systems.framework import LeafSystem, BasicVector, Context

from brom_drake.utils import AddMultibodyTriad

class ShowMeSystem(LeafSystem):
    """
    **Description**
    
    This class defines a LeafSystem that will force a specific model
    (model_index) in a MultibodyPlant to hold the joint positions
    provided to it as input. 
    
    The object may or may not be welded to the world frame.

    **Block Diagram**

    The ShowMeSystem block diagram is as follows: ::

                                    |---------------|
                                    |               |
        desired_joint_positions --->| ShowMeSystem  |---> measured_joint_positions
                                    |               |
                                    |---------------|

    **Parameters**

    plant : MultibodyPlant
        The MultibodyPlant containing the model to be shown.

    model_index : ModelInstanceIndex
        The ModelInstanceIndex of the model to be shown.

    desired_joint_positions : np.ndarray
        The desired joint positions for the model, if any.

    """
    def __init__(
        self,
        plant: MultibodyPlant,
        model_index: ModelInstanceIndex,
        desired_joint_positions: np.ndarray,
        **kwargs,
    ):
        LeafSystem.__init__(self)

        # Constants
        self.plant = plant
        self.model_index = model_index

        # TODO(kwesi): Do we need any ports?
        # Define Input Ports
        n = len(desired_joint_positions)
        self.desired_joint_positions_port = self.DeclareVectorInputPort(
            "desired_joint_positions",
            BasicVector(n),
        )

        # Define Output Ports
        self.DeclareVectorOutputPort(
            "measured_joint_positions",
            BasicVector(n),
            self.SetModelJointPositions,
            {self.time_ticket()}    # indicate that this doesn't depend on any inputs,
        )                           # but should still be updated each timestep

        # Create plant context
        self.mutable_plant_context = None

        # Define the name of the system
        self.set_name(f"ShowMeSystem_for_{plant.GetModelInstanceName(model_index)}")

    def SetModelJointPositions(self, context: Context, output: BasicVector):
        """
        **Description**
        
        This method will set the joint positions of the model to the desired joint positions.
        
        **Parameters**
        
        context: Context
            The context of the LeafSystem at the time that this callback is triggered.

        output: BasicVector
            The output vector to be populated with the current joint positions.
        """
        # Setup

        # Get desired joint positions
        pose_as_vec = self.desired_joint_positions_port.Eval(context)

        # Set the joint positions
        try:
            self.plant.SetPositions(
                self.mutable_plant_context,
                self.model_index,
                pose_as_vec,
            )
        except Exception as e:
            print(e)
            raise ValueError(
                f"Could not set the joint positions; this is most likely a mismatch between:\n" + \
                f"- Number of joint positions in command: {len(pose_as_vec)}\n" + \
                f"- Number of joint positions in model: {self.plant.num_positions(self.model_index)}."
            )

        self.plant.SetVelocities(
            self.mutable_plant_context,
            self.model_index,
            np.zeros(self.plant.num_velocities(self.model_index)),
        )

        output.SetFromVector(pose_as_vec)




from dataclasses import dataclass
import os
from pathlib import Path
import shutil
from typing import Union

# Internal Imports
from brom_drake.directories import DEFAULT_BROM_MODELS_DIR

@dataclass
class DrakeReadyURDFConverterFileManager:
    """
    Description
    -----------
    A dataclass that specifies how the DrakeReadyURDFConverter
    should manage files.
    """
    original_urdf_path: Path
    models_directory: Path = Path(DEFAULT_BROM_MODELS_DIR)
    log_file_name: str = None
    _output_urdf_path: Union[str, Path] = None

    def clean_up_models_dir(self):
        """
        Description
        -----------
        This method will clean up the models directory.
        """
        # Make sure that the models directory exists
        os.makedirs(self.models_directory, exist_ok=True) # TODO: Can we just use Pathlib here to check for existence?
        shutil.rmtree(self.models_directory, ignore_errors=True)

    def output_file_directory(self) -> Path:
        """
        Description
        -----------
        Returns the directory where the output URDF file will be saved.
        If the output_urdf_file_path is not specified, it defaults to the
        models_directory.

        Returns
        -------
        Path
            The directory where the output URDF file will be saved.
        """
        return self.output_urdf_path.parent
    
    def file_path_in_context_of_urdf_dir(self, file_path: Union[str, Path]) -> Path:
        """
        Description
        -----------
        This method returns the given file_path's Path when taken into context
        of where the urdf lies.

        Parameters
        ----------
        file_path: str or Path
            The file path to convert

        Returns
        -------
        Path
            The file path with added context of the INPUT URDF file.
        """
        # Setup
        if type(file_path) == str:
            file_path = Path(file_path)

        original_urdf_directory = self.original_urdf_path.parent
        return original_urdf_directory / file_path

    def output_urdf_file_name(self) -> str:
        """
        Description
        -----------
        Creates the output filename based on the original filename.

        Returns
        -------
        str
            The name of the output URDF file
        """
        # Setup
        original_file_path = self.original_urdf_path
        output_urdf_file_path = self._output_urdf_path

        if output_urdf_file_path is None: # If the output file path was not given.
            # Use Pathlib to extract the filename, if it exists
            original_name = original_file_path.name
            original_name = original_name.replace(".urdf", "")

            return f"{original_name}.drake.urdf"
        else:
            return Path(output_urdf_file_path).name

    @property
    def output_urdf_path(self) -> Path:
        """
        Description
        -----------
        Returns the path where the output URDF file will be saved.
        If the output_urdf_file_path is not specified, it defaults to the
        models_directory with the same name as the original URDF file.

        Returns
        -------
        Path
            The path where the output URDF file will be saved.
        """
        # Setup
        original_urdf_path = self.original_urdf_path

        # Algorithm
        if self._output_urdf_path is not None:
            if type(self._output_urdf_path) == str:
                self._output_urdf_path = Path(self._output_urdf_path)

            return self._output_urdf_path
        else:
            original_urdf_filename = original_urdf_path.name
            new_subdirectory_name = original_urdf_filename.replace(".urdf", "")
            return self.models_directory / new_subdirectory_name / self.output_urdf_file_name()

    


        



from enum import IntEnum

class AttemptGraspWithPuppeteerWristPhase(IntEnum):
    """
    *Description*
    
    An enum describing the different phases of the attempt grasp
    with puppeteer wrist production.

    Options are:
        kObjectSettlingOnFloor
        kGripperApproach
        kGripperClosing
        kObjectSettlingInGrasp
        kFloorDrop

    *Usage*

    .. code:: python

        from brom_drake.productions.debug.grasping.attempt_grasp.with_puppeteer_wrist.phases import AttemptGraspWithPuppeteerWristPhase
        # or
        # from brom_drake.all import AttemptGraspWithPuppeteerWristPhase

        # Selecting the gripper approach phase
        phase = AttemptGraspWithPuppeteerWristPhase.kGripperApproach
    """
    kObjectSettlingOnFloor = 1
    kGripperApproach = 2
    kGripperClosing = 3
    kObjectSettlingInGrasp = 4
    kFloorDrop = 5


"""
The file defining the ArmControlMode enum.
"""

from enum import IntEnum

class ArmControlMode(IntEnum):
    """
    *Description*

    An enum describing the multiple control types that can be used for an
    arm in Brom.

    Options are:
        kJoint or
        kEndEffector

    *Usage*
    
    .. code:: python

        from brom_drake.control.arms.arm_control_mode import ArmControlMode
        # or
        # from brom_drake.all import ArmControlMode
    
        # Selecting the joint control mode
        arm_control_mode = ArmControlMode.kJoint
    """
    kJoint = 0
    kEndEffector = 1


from dataclasses import dataclass
from pydrake.all import (
    ModelInstanceIndex,
    Joint,
    JointActuator
)
from typing import List

@dataclass
class PuppeteerJointSignature:
    joint: Joint
    joint_actuator: JointActuator

@dataclass
class AllJointSignatures:
    prismatic: List[PuppeteerJointSignature]
    revolute: List[PuppeteerJointSignature]

    @property
    def n_joints(self) -> int:
        return len(self.prismatic) + len(self.revolute)

@dataclass
class PuppetSignature:
    name: str
    model_instance_index: ModelInstanceIndex
    prismatic_ghost_bodies: List[ModelInstanceIndex]
    revolute_ghost_bodies: List[ModelInstanceIndex]
    joints: AllJointSignatures

    @property
    def all_joint_actuators(self) -> List[JointActuator]:
        return [sig.joint_actuator for sig in self.joints.prismatic + self.joints.revolute]

    @property
    def all_models(self) -> List[ModelInstanceIndex]:
        return self.prismatic_ghost_bodies + self.revolute_ghost_bodies + [self.model_instance_index]

    @property
    def n_joints(self) -> int:
        return self.joints.n_joints
    
    @property
    def n_models(self) -> int:
        return len(self.all_models)



from brom_drake.motion_planning.systems.open_loop_dispensers.open_loop_plan_dispenser import (
    OpenLoopPlanDispenser
)
from brom_drake.motion_planning.systems.open_loop_dispensers.pose import (
    OpenLoopPosePlanDispenser
)
from .prototypical_planner import PrototypicalPlannerSystem
from .state_of_plan_in_memory import StateOfPlanInMemory

__all__ = [
    "OpenLoopPlanDispenser",
    "OpenLoopPosePlanDispenser",
    "PrototypicalPlannerSystem",
    "StateOfPlanInMemory",
]


from brom_drake.productions.motion_planning.offline.kinematic.shelf import ShelfPlanning1
from brom_drake.productions.motion_planning.offline.kinematic.chem_lab1 import ChemLab1
from brom_drake.productions.motion_planning.offline.dynamic.chem_lab2 import ChemLab2

__all__ = [
    "ShelfPlanning1",
    "ChemLab1",
    "ChemLab2",
]


from dataclasses import dataclass, field
from typing import List

from brom_drake.productions.types.base.configuration import Configuration as BaseConfiguration 

@dataclass
class Configuration:
    base: BaseConfiguration = field(default_factory=BaseConfiguration) #By default, use the standard script
    show_collision_geometries: bool = False
    target_body_on_gripper: str = None # The "Gripper" Frame that we use to define X_GripperObject
    gripper_color: List[float] = None
    show_gripper_base_frame: bool = False

    @property
    def meshcat_port_number(self):
        return self.base.meshcat_port_number

    @meshcat_port_number.setter
    def meshcat_port_number(self, new_value: int):
        self.base.meshcat_port_number = new_value


from .configuration import Configuration as RigidTransformToVectorSystemConfiguration
from .system import RigidTransformToVectorSystem

__all__ = [
    "RigidTransformToVectorSystem",
    "RigidTransformToVectorSystemConfiguration",
]


"""
bidirectional_connect.py
Description
-----------
This file contains an algorithm for performing bidirectional RRT-Connect
based search.
"""

from dataclasses import dataclass
import networkx as nx
import numpy as np
from pydrake.all import (
    ModelInstanceIndex,
    MultibodyPlant,
    SceneGraph,
)
from typing import Callable, Tuple

# Internal Imports
from brom_drake.motion_planning.algorithms.motion_planner import MotionPlanner
# TODO(Kwesi): Introduce planning node object

# Define config dataclass
@dataclass(frozen=True)
class BiRRTConnectSamplingProbabilities:
    """
    A dataclass that defines the sampling probabilities for a bidirectional RRT planner.
    """
    sample_goal_tree: float = 0.10
    sample_opposite_tree: float = 0.10
    

@dataclass
class BidirectionalRRTConnectPlannerConfig:
    """
    A dataclass that defines the configuration for a bidirectional RRT planner.
    """
    convergence_threshold: float = 1e-3
    max_tree_nodes: int = int(1e5)
    probabilities: BiRRTConnectSamplingProbabilities = BiRRTConnectSamplingProbabilities()
    random_seed: int = 23
    steering_step_size: float = 0.1
    debug: bool = False

class BidirectionalRRTConnectPlanner(MotionPlanner):
    def __init__(
        self,
        robot_model_idx: ModelInstanceIndex,
        plant: MultibodyPlant,
        scene_graph: SceneGraph,
        config: BidirectionalRRTConnectPlannerConfig = None,
    ):
        # Input Processing
        if config is None:
            config = BidirectionalRRTConnectPlannerConfig()

        # Setup
        self.config = config

        # Use the parent class constructor
        super().__init__(
            robot_model_idx, plant, scene_graph, 
            random_seed=self.config.random_seed,
        )

    def add_new_node_to(
        self,
        target_tree: nx.DiGraph,
        prev_node: nx.classes.reportviews.NodeView,
        q_new: np.ndarray,
        tree_is_goal: bool,
    ):
        """
        Description
        -----------
        This function adds a new node to the RRT.

        Arguments
        ---------
        target_tree: nx.DiGraph
            The RRT to add the node to.
        q_new: np.ndarray
            The new configuration to add to the RRT.
        """
        # Setup
        n_nodes = target_tree.number_of_nodes()

        # Add the new node to the RRT
        target_tree.add_node(n_nodes, q=q_new)
        if tree_is_goal:
            target_tree.add_edge(
                target_tree.number_of_nodes()-1,
                prev_node,
            )
        else:
            target_tree.add_edge(
                prev_node,
                target_tree.number_of_nodes()-1,
            )

    def connect(
        self,
        nearest_node_view: nx.classes.reportviews.NodeView,
        q_target: np.ndarray,
        rrt: nx.DiGraph,
        tree_is_goal: bool,
        collision_check_fcn: Callable[[np.ndarray], bool] = None,
        target_config_already_exists: bool = False,
    ) -> Tuple[np.ndarray, bool]:
        """
        Description
        -----------
        This function connects the RRT from the nearest node to the target configuration.

        Arguments
        ---------
        target_config_already_exists: bool
            Whether or not the target configuration already exists in the combined RRT.
        """
        # Input Processing
        if collision_check_fcn is None:
            collision_check_fcn = self.check_collision_in_config

        # Setup
        nearest_node = rrt.nodes[nearest_node_view]
        debug = self.config.debug

        # Steer from nearest node to random configuration
        q_current = nearest_node['q']
        last_node_view = nearest_node_view

        n_loops = 0
        reached_new = False
        while not np.isclose(q_current, q_target).all():
            # Create next node
            q_new, reached_new = self.steer(q_current, q_target)

            # Check for collisions
            if collision_check_fcn(q_new):
                break

            # Check if we reached the target configuration
            if reached_new and target_config_already_exists:
                break

            # If the next point is collision free, then add new node to the tree
            self.add_new_node_to(
                rrt,
                last_node_view,
                q_new,
                tree_is_goal=tree_is_goal,
            )

            # Prepare for next iteration
            last_node_view = list(rrt.nodes)[-1]
            q_current = q_new

            n_loops += 1

        if debug:
            print(f"Number of loops in connection: {n_loops}")
            print(f"RRT Size: {rrt.number_of_nodes()}")

        return q_current, reached_new # Return the last configuration we visited

    def plan(
        self,
        q_start: np.ndarray,
        q_goal: np.ndarray,
        collision_check_fcn: Callable[[np.ndarray], bool] = None,
    ) -> Tuple[nx.DiGraph, int]:
        """
        Description
        -----------
        This function plans a path from q_start to q_goal using the Bidirectional
        RRT algorithm.

        Returns
        -------
        nx.DiGraph
            The RRT that was created during the planning process.
        int
            The index of the goal node in the RRT.
        """
        # Input Processing
        q_start = np.array(q_start)
        q_goal = np.array(q_goal)
        if q_start.shape[0] != self.dim_q or q_goal.shape[0] != self.dim_q:
            raise ValueError(
                f"Start configuration shape ({q_start.shape}) and goal configuration " +
                f"shape ({q_goal.shape}) must match the dimension of the robot ({self.dim_q})."
            )

        if collision_check_fcn is None:
            collision_check_fcn = self.check_collision_in_config

        # Setup
        rrt_start = nx.DiGraph()
        rrt_goal = nx.DiGraph()
        rrt_start.add_node(0, q=q_start)
        rrt_goal.add_node(0, q=q_goal)

        prob_sample_goal_tree = self.config.probabilities.sample_goal_tree
        prob_sample_opposite_tree = self.config.probabilities.sample_opposite_tree
        max_tree_nodes = self.config.max_tree_nodes

        # Initialize the trees
        n_nodes_start = 1
        n_nodes_goal = 1
        sample_from_goal_tree = None
        while (n_nodes_start + n_nodes_goal) <= max_tree_nodes:
            # Decide whether to sample from the start or goal tree
            sample_from_goal_tree = np.random.rand() < prob_sample_goal_tree
            if sample_from_goal_tree: # Sample from the goal tree
                current_tree = rrt_goal
            else: # Sample from the start tree
                current_tree = rrt_start

            # Choose whether or not to sample from the opposite tree or randomly
            if np.random.rand() < prob_sample_opposite_tree:
                combined_tree, connected_two_trees = self.steer_towards_tree(
                    rrt_start,
                    rrt_goal,
                    current_tree_is_goal=sample_from_goal_tree,
                    collision_check_fcn=collision_check_fcn,
                )
                if connected_two_trees:
                    return combined_tree, rrt_start.number_of_nodes()

            else:
                # Sample from a random configuration
                q_random = self.sample_random_configuration()

                # Find the nearest node in the tree and steer towards it
                nearest_node_view, nearest_node_idx = self.sample_nearest_in_tree(current_tree, q_random)
                q_new, reached_new = self.connect(
                    nearest_node_view,
                    q_random,
                    current_tree,
                    tree_is_goal=sample_from_goal_tree,
                    collision_check_fcn=collision_check_fcn,
                )
                # q_new, reached_new = self.steer(
                #     current_tree.nodes[nearest_node_view]['q'],
                #     q_random,
                # )

                # # Check if the new configuration is in collision
                # if collision_check_fcn(q_new):
                #     continue

                # # If the configuration is not in collision,
                # # then add it to the current tree
                # self.add_new_node_to(
                #     current_tree,
                #     nearest_node_view,
                #     q_new,
                #     tree_is_goal=sample_from_goal_tree,
                # )

            # Update the number of nodes in the appropriate tree
            n_nodes_goal = rrt_goal.number_of_nodes()
            n_nodes_start = rrt_start.number_of_nodes()

        # If we exit the loop without finding a path to the goal,
        # return the RRT and indicate failure
        print("Max iterations reached without finding a path to the goal.")
        return nx.disjoint_union(rrt_start, rrt_goal), -1

    def sample_from_tree(
        self,
        rrt: nx.DiGraph,
    ) -> nx.classes.reportviews.NodeView:
        """
        Description
        -----------
        This function samples a configuration from the RRT.
        """
        # Setup
        n_tree = rrt.number_of_nodes()

        # Select a random number from the range [0, n_tree)
        random_index = np.random.randint(0, n_tree)
        for ii, node_ii in enumerate(rrt.nodes):
            if ii == random_index:
                return node_ii

        raise ValueError(
            f"Random index ({random_index}) not found in RRT with {n_tree} nodes."
        )
    
    def sample_nearest_in_tree(
        self,
        rrt: nx.DiGraph,
        q_random: np.ndarray,
    ) -> Tuple[nx.classes.reportviews.NodeView, float]:
        """
        Description
        -----------
        This function samples a configuration from the RRT.

        Arguments
        ---------
        rrt: nx.DiGraph
            The RRT to sample from.
        q_random: np.ndarray
            The random configuration to sample from the RRT.

        Returns
        -------
        node_view: nx.classes.reportviews.NodeView
            A reference to the nearest node in the RRT.
            Use this to access the node as follows rrt.nodes[node_view].
        distance: float
            The distance from the random configuration to the nearest node in the RRT.
            
        """
        # Setup

        # Search through the tree to find the node that is nearest to the random configuration
        min_distance = float('inf')
        nearest_node_view = None
        for node in rrt.nodes:
            distance = np.linalg.norm(rrt.nodes[node]['q'] - q_random)
            if distance < min_distance:
                min_distance = distance
                nearest_node_view = node

        return nearest_node_view, min_distance
    
    def steer(
        self,
        q_current: np.ndarray,
        q_target: np.ndarray,
    ) -> Tuple[np.ndarray, bool]:
        """
        Description
        -----------
        This function steers the RRT from the start configuration to the goal configuration.

        Arguments
        ---------
        q_current: np.ndarray
            The current configuration of the robot that we wish to steer from.
        q_target: np.ndarray
            The target configuration of the robot that we wish to steer to.

        Returns
        -------
        q_new: np.ndarray
            The new configuration of the robot after steering.
        reached_config: bool
            Whether or not we reached the target configuration with our step size or not.
        """
        # Setup
        step_size = self.config.steering_step_size

        # Calculate the direction vector
        direction = q_target - q_current
        distance = np.linalg.norm(direction)

        # Either:
        # 1. Move a full step towards the random configuration
        # 2. If the distance is less than the step size, return the random configuration

        if distance < step_size:
            return q_target, True
        
        return q_current + step_size * (direction / distance), False
    
    def steer_towards_tree(
        self,
        rrt_start: nx.DiGraph,
        rrt_goal: nx.DiGraph,
        current_tree_is_goal: bool,
        collision_check_fcn: Callable[[np.ndarray], bool] = None,
    ) -> Tuple[nx.DiGraph, bool]:
        """
        Description
        -----------
        This function steers the RRT from the start configuration to the goal configuration.

        Arguments
        ---------

        Returns
        -------
        bool
            Whether or not we reached the target configuration with our step size or not.
        """
        # Setup
        debug = self.config.debug
        current_tree = rrt_goal if current_tree_is_goal else rrt_start

        if collision_check_fcn is None:
            collision_check_fcn = self.check_collision_in_config

        # Sample a node in the current tree
        current_node_view = self.sample_from_tree(current_tree)
        q_current = current_tree.nodes[current_node_view]['q']

        # Sample from the opposite tree
        opposite_tree = rrt_start if current_tree_is_goal else rrt_goal
        node_idx_in_opposite_tree, min_distance = self.sample_nearest_in_tree(opposite_tree, q_current)
        
        node_in_opposite_tree = opposite_tree.nodes[node_idx_in_opposite_tree]
        sampled_config = node_in_opposite_tree['q']

        # Steer from the current configuration to the sampled one
        # q_new, reached_new = self.steer(q_current, sampled_config)
        q_new, reached_new = self.connect(
            current_node_view,
            sampled_config,
            current_tree,
            tree_is_goal=current_tree_is_goal,
            collision_check_fcn=collision_check_fcn,
            target_config_already_exists=True,
        )

        # Check if the new configuration is in collision
        if collision_check_fcn(q_new):
            return None, False

        if reached_new:
            combined_rrt = nx.disjoint_union(rrt_start, rrt_goal)
            # add edge between the two trees
            if current_tree_is_goal:
                if debug:
                    print(f"Adding edge between {node_idx_in_opposite_tree} and {rrt_start.number_of_nodes() + int(current_node_view)}")

                combined_rrt.add_edge(
                    node_idx_in_opposite_tree,
                    rrt_start.number_of_nodes() + int(current_node_view),
                )
            else:
                if debug:
                    print(f"Adding edge between {current_node_view} and {rrt_start.number_of_nodes() + node_idx_in_opposite_tree}")

                combined_rrt.add_edge(
                    int(current_node_view),
                    rrt_start.number_of_nodes() + node_idx_in_opposite_tree,
                )
            return combined_rrt, True
        else:
            # If we did not reach the goal,
            # add the new configuration to the current tree
            self.add_new_node_to(
                current_tree,
                current_node_view,
                q_new,
                tree_is_goal=current_tree_is_goal,
            )

        # We did not finish, so do not return the combined RRT
        # and tell the caller that we did not finish
        return None, False


from .abstract_list_selection_system import AbstractListSelectionSystem
from .abstract_port_switch_system import AbstractPortSwitch
from .pose_composition import PoseCompositionSystem

__all__ = [
    "AbstractListSelectionSystem",
    "AbstractPortSwitch",
    "PoseCompositionSystem",
]


import os
from pathlib import Path
from typing import List, Tuple, Callable

import networkx as nx
import numpy as np
from pydrake.all import (
    RotationMatrix, Quaternion, RollPitchYaw,
    Solve, SolutionResult,
    InverseKinematics, ModelInstanceIndex,
)
from pydrake.common.value import AbstractValue
from pydrake.math import RigidTransform
from pydrake.multibody.parsing import Parser
from pydrake.multibody.plant import MultibodyPlant
from pydrake.systems.framework import Diagram, Context
from pydrake.systems.primitives import ConstantVectorSource, ConstantValueSource

# Internal Imports
from brom_drake.motion_planning.systems.prototypical_planner import PrototypicalPlannerSystem
from brom_drake.productions.types import BaseProduction
from brom_drake.productions.roles import kKinematicMotionPlanner, Role
from brom_drake.file_manipulation.urdf.shapes.sphere import SphereDefinition
from brom_drake.file_manipulation.urdf.simple_writer.urdf_definition import SimpleShapeURDFDefinition
from brom_drake.utils import Performer, MotionPlan


class KinematicMotionPlanningProduction(BaseProduction):
    """
    *Description*

    This class defines a basic kinematic motion planning production.

    It includes the ability to define start and goal configurations or poses,
    and to add them to the plant as visual elements.

    .. important::

        This production assumes that either the start pose or configuration is defined.
        (One can be inferred from the other using inverse kinematics, if needed.)
        Similarly, either the goal pose or configuration must be defined.

    *Parameters*

    start_configuration: np.ndarray, optional
        The starting configuration of the robot, by default None.

    start_pose: RigidTransform, optional
        The starting pose of the robot, by default None.

    goal_configuration: np.ndarray, optional
        The goal configuration of the robot, by default None.

    goal_pose: RigidTransform, optional
        The goal pose of the robot, by default None.
    """
    def __init__(
        self,
        start_configuration: np.ndarray = None,
        start_pose: RigidTransform = None,
        goal_configuration: np.ndarray = None,
        goal_pose: RigidTransform = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        # Note the following objects will exist:
        # - self.plant
        # - self.scene_graph

        # Start and Goal Configurations
        self._start_config = start_configuration
        self._goal_config = goal_configuration

        # Start and Goal Poses
        self._start_pose = start_pose
        self._goal_pose = goal_pose

        # Create placeholder for the robot model index
        self.robot_model_idx_ = None

    def add_supporting_cast(self):
        """
        *Description*
        
        Create illustrations for the start and goal poses to the builder.

        .. important::

            The child classes of this base class should add their
            own logic to their own add_supporting_cast() method, and
            then call this method via super().add_supporting_cast().

            This method is only meant to add the start and goal visualizations
            to the plant.
        
        .. warning::

            We recommend that the plant NOT be finalized before calling this method.
            This is because this method adds visual elements to the plant.
        
            
        *Usage*

        To demonstrate how to use this method in a child class, see the
        following example: ::

            class MyKinematicMotionPlanningProduction(KinematicMotionPlanningProduction):
                # Some code here...

                def add_supporting_cast(self):
                    # First, call the parent class's method to add start and goal visuals
                    super().add_supporting_cast()

                    # Then, add the robot and anything else
                    self.add_robot_to_plant(self.plant)

                    # Finally, finalize the plant
                    self.plant.Finalize()

                    # Create systems which share the pose where the start and goal
                    # are located.
                    self.add_start_and_goal_sources_to_builder()



        """
        # Add visual elements for the start and goal poses
        self.add_start_and_goal_to_plant(self.plant)

    def add_robot_source_system(self):
        """
        *Description*
        
        This method adds a source for providing the motion planner
        with the model index for the robot that we are trying to control.
        """
        # Setup

        # Create AbstractValueSource
        robot_source_system = ConstantValueSource(
            AbstractValue.Make(self.robot_model_idx_)
        )
        robot_source_system.set_name("robot_model_index_source")

        self.builder.AddSystem(robot_source_system)

    def add_start_and_goal_to_plant(self, plant: MultibodyPlant):
        """
        *Description*
        
        Add 3D visualizations of the pose where the start and goal
        are for the current production.

        .. tip::

            Feel free to overwrite this if you have specific sizes of the start and goal in mind,
            or if you want to add more than just spheres, etc.
        
        *Parameters*
        
        plant: pydrake.multibody.plant.MultibodyPlant
            The plant to add the start and goal to.
        """

        # Create a sphere for the start
        start_sphere_defn = SimpleShapeURDFDefinition(
            name="start_sphere",
            shape=SphereDefinition(radius=0.1),
            create_collision=False,
            color=np.array([0, 0.3, 0.8, 0.3]),
        )
        start_sphere_urdf_location = Path("brom/models/spheres/start.urdf")
        os.makedirs(start_sphere_urdf_location.parent, exist_ok=True)

        start_sphere_defn.write_to_file(start_sphere_urdf_location)

        # Load the start sphere into the plant and rigidly attach it at the start_pose
        start_sphere_model_idx = Parser(plant).AddModels(str(start_sphere_urdf_location))[0]
        plant.WeldFrames(
            plant.world_frame(),
            plant.GetFrameByName("start_sphere_base_link", start_sphere_model_idx),
            self.start_pose,
        )

        # Create a sphere for the goal
        goal_sphere_defn = SimpleShapeURDFDefinition(
            name="end_sphere",
            shape=SphereDefinition(radius=0.1),
            create_collision=False,
            color=np.array([0, 0.8, 0.3, 0.3]),
        )
        goal_sphere_urdf_location = Path("brom/models/spheres/end.urdf")
        os.makedirs(goal_sphere_urdf_location.parent, exist_ok=True)

        goal_sphere_defn.write_to_file(goal_sphere_urdf_location)

        # Load the start sphere into the plant and rigidly attach it at the start_pose
        goal_sphere_model_idx = Parser(plant).AddModels(str(goal_sphere_urdf_location))[0]
        plant.WeldFrames(
            plant.world_frame(),
            plant.GetFrameByName(f"{goal_sphere_defn.name}_base_link", goal_sphere_model_idx),
            self.goal_pose,
        )

    def add_start_and_goal_sources_to_builder(self):
        """
        *Description*
        
        This method will add the start and goal sources to the builder.

        .. warning::

            The plant must be finalized before calling this method.
        """
        # Input Processing
        assert self.plant.is_finalized(), "Plant must be finalized before adding start and goal sources."

        # Add the start and goal poses to the builder
        self.add_start_source_system()
        self.add_goal_source_system()

    def add_start_source_system(self):
        """
        *Description*

        Add a source system for the start pose.
        This will be a ConstantVectorSource that outputs the start configuration
        in the diagram.
        """
        # Setup

        # Adds a source system for the start pose to the builder
        start_source_system = ConstantVectorSource(self.start_configuration)
        start_source_system.set_name("start_configuration")

        self.builder.AddSystem(start_source_system)

    def add_goal_source_system(self):
        """
        *Description*
        
        Add a source system for the goal pose.
        
        This will be a ConstantVectorSource that outputs the goal configuration
        in the diagram.
        """
        # Setup

        # Adds a source system for the goal pose to the builder
        goal_source_system = ConstantVectorSource(self.goal_configuration)
        goal_source_system.set_name("goal_configuration")

        self.builder.AddSystem(goal_source_system)

    def add_main_cast(
        self,
        cast: List[Tuple[Role, Performer]] = [],
    ):
        """
        *Description*
        
        This small modification to the normal add_main_cast()
        function includes a call to create the optional inputs if needed.
        
        *Parameters*

        cast: List[Tuple[Role, Performer]], optional
            The cast to add to the production, by default [].
        """
        # Setup

        # Default
        super().add_main_cast(cast=cast)

        # Create optional outputs, if needed
        for role_ii, performer_ii in cast:
            self.create_optional_outputs_if_necessary(
                role_ii, performer_ii
            )

    def create_optional_outputs_if_necessary(
        self,
        role: Role, # This role will always be the Motion Planner, won't it?
        performer: Performer,
    ):
        """
        Description
        -----------
        This method checks to see if all optional outputs were connected
        for the given role. If all of them were, then we do nothing.
        If any of the optional outputs were NOT connected,
        then we add a dummy value to replace it.

        Requirements
        ------------
        This method should only be called after the full production has
        been built. Otherwise, this may have unexpected behavior.

        :return:
        """
        # Setup

        # Iterate through all OPTIONAL OUTPUTS
        # for assignment_ii in role.port_assignments:
        #     # Should we investigate this port?
        #     investigate_this_assignment = assignment_ii.pairing_type == PairingType.kOutput
        #     investigate_this_assignment = investigate_this_assignment and \
        #                                   (not assignment_ii.is_required)
        #
        #     if not investigate_this_assignment:
        #         continue # Skip this assignment value if it doesn't satisfy this
        #
        #     # Otherwise, check to see if we have the desired output port
        #     if performer.HasOutputPort(assignment_ii.performer_port_name):
        #         continue # If performer has the output port, then we should have already connected it; skip.
        #
        #     # If performer DOES NOT HAVE

        # TODO(kwesi): Maybe move above, general code to somewhere else?
        if role.name != "OfflineMotionPlanner":
            raise ValueError(
                f"Expected role for OfflineMotionPlanning Production to be \"OfflineMotionPlanner\";" +
                f"received {role.name}"
            )

        # Check to see if performer has the LAST assignment in the motion planning role
        last_assignment = role.port_assignments[-1]
        if last_assignment.performer_port_name != "plan_is_ready":
            raise ValueError(
                f"Expected last port assignment to be \"plan_is_ready\" assignment, but" +
                f"received an assignment with performer_port_name \"{last_assignment.performer_port_name}\""
            )

        if performer.HasOutputPort(last_assignment.performer_port_name):
            return # Do nothing; performer should already be connected

        # If the performer does not have plan_is_ready port, then
        # let's create a dummy value and connect it to the right place.
        plan_ready_source = ConstantValueSource(
            AbstractValue.Make(True),
        )
        self.builder.AddSystem(plan_ready_source)

        # Find system we want to connect it to
        systems_list = last_assignment.find_any_matching_input_targets(self.builder)
        assert len(systems_list) == 1, \
            f"Expected 1 system to have port \"{last_assignment.external_target_name}\"," + \
            f" but found {len(systems_list)} systems with that output port."
        # TODO(kwesi): Perhaps move more of these error assertions to a separate file?

        self.builder.Connect(
            plan_ready_source.get_output_port(),
            systems_list[0].GetInputPort(last_assignment.external_target_name)
        )

    def define_pose_ik_problem(
        self,
        pose_WorldTarget: RigidTransform,
        target_frame_name: str,
        eps0: float = 2.5e-2,
        orientation_cost_scaling: float = 0.25,
    ) -> InverseKinematics:
        """
        Description
        -----------
        Sets up the inverse kinematics problem for the start pose
        input to theis function.
        :return:
        """
        # Setup

        # Create IK Problem
        ik_problem = InverseKinematics(self.plant)

        # Add Pose Target
        ik_problem.AddPositionConstraint(
            self.plant.world_frame(),
            pose_WorldTarget.translation(),
            self.plant.GetFrameByName(target_frame_name),
            (- np.ones((3,)) * eps0).reshape((-1, 1)),
            (+ np.ones((3,)) * eps0).reshape((-1, 1)),
        )

        # TODO(kwesi): Add OrientationCosntraint
        # ik_problem.AddOrientationConstraint(
        #     self.plant.world_frame(),
        #     RotationMatrix(Quaternion(input_pose_vec[3:]).rotation()),
        #     self.plant.GetFrameByName("ft_frame"),
        #     RotationMatrix.Identity(),
        #     0.25,
        # )

        ik_problem.AddOrientationCost(
            self.plant.world_frame(),
            pose_WorldTarget.rotation(),
            self.plant.GetFrameByName(target_frame_name),
            RotationMatrix.Identity(),
            orientation_cost_scaling,
        )

        return ik_problem
    
    def easy_cast_and_build(
        self,
        planning_algorithm: Callable[
            [np.ndarray, np.ndarray, Callable[[np.ndarray], bool]],
            Tuple[MotionPlan, bool],
        ],
        with_watcher: bool = True,
    ) -> Tuple[Diagram, Context]:
        """
        Description
        -----------
        This function is used to easily cast and build the production.
        :param planning_algorithm: The algorithm that we will use to
        plan the motion.
        :param with_watcher: A Boolean that determines whether to add a watcher to the diagram.
        :return:
        """
        # Setup

        # Add all elements to the builder
        self.add_supporting_cast()
        print("added supporting cast")

        # Create a planner from the algorithm
        prototypical_planner = PrototypicalPlannerSystem(
            self.plant, self.scene_graph,
            planning_algorithm,
            robot_model_idx=self.robot_model_index,
        )

        # Cast the production using the prototypical planner
        planner_role = self.suggested_roles()[0]

        # Fulfill each role-performer pair in the casting_call list
        self.fill_role(planner_role, prototypical_planner)
        print("filled role")

        self.create_optional_outputs_if_necessary(
            planner_role, prototypical_planner
        )

        # Build
        return self.build_production(with_watcher=with_watcher)

    @property
    def goal_configuration(self) -> np.ndarray:
        """
        Get the goal pose. This should be defined by the subclass.
        :return:
        """
        if self._goal_config is not None:
            return self._goal_config
        elif self._goal_pose is not None:
            # Use the goal pose to get the goal configuration
            # Using the IK solver (potentially buggy because default ik problem ignores obstacles)
            return self.solve_pose_ik_problem(self._goal_pose)
        else:
            raise NotImplementedError(
                "This function should be implemented by the subclass."
            )

    @property
    def goal_pose(self) -> RigidTransform:
        """
        Get the goal pose. This should be defined by the subclass.
        :return:
        """
        if self._goal_pose is not None:
            return self._goal_pose
        elif self._goal_config is not None:
            # The user would need to define a forward kinematics function,
            # to convert the goal configuration to a goal pose.
            raise NotImplementedError(
                "It looks like you are trying to get the goal pose from the goal configuration.\n" +
                "This is not implemented (yet!) because it would require solving the forward kinematics" +
                "problem for your specific scene.\n" +
                "Please implement your own goal_pose() function with a custom forward kinamtics function" +
                " (see ChemLab2's forward kinematics function) to convert the goal configuration to a goal pose."
            )
            #TODO(kwesi): Is there a way to do this without knowing about the user's robot?
        else:
            raise NotImplementedError(
                "This function should be implemented by the subclass."
            )

    @property
    def robot_model_index(self) -> ModelInstanceIndex:
        """
        Get the robot model index. This should be defined by the subclass.
        :return:
        """
        if self.robot_model_idx_ is not None:
            return self.robot_model_idx_
        else:        
            raise NotImplementedError(
                "This function should be implemented by the subclass."
            )

    @property
    def start_configuration(self) -> np.ndarray:
        """
        Description:
            This function returns the start and goal poses as a single array.
        """
        if self._start_config is not None:
            return self._start_config
        elif self._start_pose is not None:
            # Use the start pose to get the start configuration
            # Using the IK solver (potentially buggy because default ik problem ignores obstacles)
            return self.solve_pose_ik_problem(self._start_pose)
        else:
            raise NotImplementedError(
                "This function should be implemented by the subclass."
            )
    
    @property
    def start_pose(self) -> RigidTransform:
        """
        Get the start pose. This should be defined by the subclass.
        :return:
        """
        if self._start_pose is not None:
            return self._start_pose
        elif self._start_config is not None:
            # The user would need to define a forward kinematics function,
            # to convert the start configuration to a start pose.
            raise NotImplementedError(
                "It looks like you are trying to get the start pose from the start configuration.\n" +
                "This is not implemented (yet!) because it would require solving the forward kinematics" +
                "problem for your specific scene.\n" +
                "Please implement your own start_pose() function with a custom forward kinamtics function" +
                " (see ChemLab2's forward kinematics function) to convert the start configuration to a start pose."
            )
            #TODO(kwesi): Is there a way to do this without knowing about the user's robot?
        else:
            raise NotImplementedError(
                "This function should be implemented by the subclass."
            )

    def suggested_roles(self):
        return [kKinematicMotionPlanner]

    def solve_pose_ik_problem(
        self,
        pose_WorldTarget: RigidTransform,
        frame_name: str = "ft_frame",
        robot_joint_names: list = None,
    ) -> np.ndarray:
        """
        Description:
        :param input_pose_vec: An n-dimensional vector
        :return:
        """
        # Input Processing
        assert self.plant is not None, "Plant must be defined before solving IK problem!"

        # Setup

        # Define Problem
        ik_problem = self.define_pose_ik_problem(
            pose_WorldTarget,
            frame_name,
        )

        # Solve problem
        ik_program = ik_problem.prog()
        ik_result = Solve(ik_program)

        assert ik_result.get_solution_result() == SolutionResult.kSolutionFound, \
            f"Solution result was {ik_result.get_solution_result()}; need SolutionResult.kSolutionFound to make RRT Plan!"

        q_solution = ik_result.get_x_val()
        # print(f"solved ik problem: {q_solution}")

        # Extract only the positions that correspond to our robot's joints
        if robot_joint_names is None:
            robot_joint_names = self.plant.GetPositionNames(self.robot_model_index, add_model_instance_prefix=True)
        
        all_joint_names = self.plant.GetPositionNames()
        q_out_list = []
        for ii, joint_name in enumerate(all_joint_names):
            if joint_name in robot_joint_names:
                q_out_list.append(q_solution[ii])

        # print(f"q_out_list: {q_out_list}")

        return np.array(q_out_list)


from dataclasses import dataclass
from pydrake.all import (
    AddMultibodyPlantSceneGraph,
    Frame,
    Meshcat,
    MeshcatVisualizer,
    MeshcatVisualizerParams,
    MultibodyPlant,
    Parser,
    RigidTransform
)
from pydrake.all import Role as DrakeRole
from typing import List

# Internal Imports
from brom_drake.file_manipulation.urdf.drakeify import drakeify_my_urdf
from brom_drake.productions.types.base.base import BaseProduction
from brom_drake.utils.model_instances import (
    get_name_of_first_body_in_urdf,
    get_name_of_all_bodies_in_urdf,
)
from brom_drake.utils.triads import AddMultibodyTriad

class BasicGraspingDebuggingProduction(BaseProduction):
    """
    Description
    -----------
    This base class is used to define several functions that are useful
    for any production that is used for basic grasping tasks.
    """
    def __init__(
        self, 
        path_to_object: str,
        path_to_gripper: str,
        X_ObjectGrasp: RigidTransform,
        meshcat_port_number: int = None,
        time_step: float = 0.001,
        target_body_on_gripper: str = None,
        gripper_color: List[float] = None,
        show_gripper_base_frame: bool = False,
        show_collision_geometries: bool = False,
    ):
        """
        Description
        -----------
        Defines the following fields for the production:
        
        Arguments
        ----------
        path_to_object: str
            A string representing the path to the object urdf file.
        path_to_gripper: str
            A string representing the path to the gripper urdf file.
        X_ObjectGrasp: RigidTransform
            The transform of the Grasp frame (i.e., where the grasp should start)
            with respect to the object being grasped.
        meshcat_port_number: int (Optional, default=None)
            An integer representing the port number for the meshcat server.
            If this is None, the meshcat server will not be started.
        time_step: float (Optional, default=0.001)
            A float representing the time step for the simulation.
        target_body_on_gripper: str (Optional, default=None)
            A string representing the name of the body on the gripper that will be used as the target frame.
            If this is None, the first body in the gripper will be used.
        gripper_color: List[float]
            A list of floats representing the color of the gripper in RGBA format.
        show_gripper_base_frame: bool
            A boolean flag that indicates whether or not to show the base frame of the gripper.
        show_collision_geometries: bool
            A boolean flag that indicates whether or not to show the collision geometries of the gripper.
        """
        # Call the base class constructor
        super().__init__()

        # Note the following objects will exist:
        # - self.plant
        # - self.scene_graph

        # Add the model to the Production
        self.path_to_object = path_to_object
        self.path_to_gripper = path_to_gripper
        self.time_step = time_step
        self.gripper_color = gripper_color
        self.show_gripper_base_frame = show_gripper_base_frame
        
        if X_ObjectGrasp is None:
            X_ObjectGrasp = RigidTransform()
        self.X_ObjectGripper = X_ObjectGrasp
        
        self.meshcat_port_number = meshcat_port_number
        self.show_collision_geometries = show_collision_geometries

        # Assign the target frame on the gripper to the variable
        if target_body_on_gripper is None:
            target_body_on_gripper = get_name_of_first_body_in_urdf(self.path_to_gripper)
        else:
            assert target_body_on_gripper in get_name_of_all_bodies_in_urdf(self.path_to_gripper), \
                f"Target body \"{target_body_on_gripper}\" not found in gripper model; Valid body names are: {self.get_all_body_names_in_gripper()}."


        self.target_body_name_on_gripper = target_body_on_gripper

        # Add Plant and Scene Graph for easy simulation
        self.plant, self.scene_graph = AddMultibodyPlantSceneGraph(
            self.builder,
            time_step=1e-3,
        )

        # Create empty variables to be filled in later.
        self.meshcat = None
        self.manipuland_index, self.manipuland_name = None, None
        self.gripper_model_index, self.gripper_model_name = None, None

    def add_gripper_to_plant(self, and_weld_to: Frame = None, with_X_WorldGripper: RigidTransform = None):
        """
        Description
        -----------
        This method will add the gripper to the plant
        and then weld it to the origin.
        """
        # Setup
        assert self.gripper_model_index is None, \
            "The gripper model index is already set. Please DO NOT add the gripper to the plant before this function."

        plant : MultibodyPlant = self.plant
        gripper_color = self.gripper_color

        show_gripper_base_frame = self.show_gripper_base_frame
        if with_X_WorldGripper is None:
            with_X_WorldGripper = RigidTransform()
        

        # Input Processing
        if gripper_color is not None:
            # Convert gripper URDF to Drake-ready URDF
            recolored_gripper_urdf = drakeify_my_urdf(
                self.path_to_gripper,
                overwrite_old_logs=True,
                log_file_name="BasicGraspingDebuggingProduction_AddManipulandToPlant_gripper.log",
                replace_colors_with=gripper_color,
            )
            self.path_to_gripper = str(recolored_gripper_urdf)

        # Add the gripper to the plant
        temp_idcs = Parser(plant=self.plant).AddModels(
            self.path_to_gripper,
        )
        assert len(temp_idcs) == 1, f"Only one model should be added; received {len(temp_idcs)}"
        self.gripper_model_index = temp_idcs[0]
        self.gripper_model_name = self.plant.GetModelInstanceName(self.gripper_model_index)

        # Draw the MultibodyTriad for the
        # - Target Frame on the Gripper
        # - Base Link of the Gripper
        target_frame = plant.GetFrameByName(self.target_body_name_on_gripper)
        gripper_base_frame = plant.GetFrameByName(
            get_name_of_first_body_in_urdf(self.path_to_gripper)
        )

        # Add the target triad to the builder
        AddMultibodyTriad(
            target_frame,
            self.scene_graph,
        )

        # Add the gripper triad to the builder
        target_is_different_from_gripper_base = target_frame.body().name() != gripper_base_frame.body().name()
        if show_gripper_base_frame and target_is_different_from_gripper_base:
            AddMultibodyTriad(
                gripper_base_frame,
                self.scene_graph,
                # scale=0.1,
            )
            

        # Weld the gripper to the manipuland
        if and_weld_to is not None:
            plant.WeldFrames(
                and_weld_to,
                gripper_base_frame,
                with_X_WorldGripper,
            )

    def add_manipuland_to_plant(self, and_weld_to: Frame = None):
        """
        Description
        -----------
        This method will add the manipuland to the plant
        and then weld it to the origin.
        """
        # Setup
        assert self.manipuland_index is None, \
            "The manipuland index is already set. Please DO NOT add the manipuland to the plant before this function."

        plant : MultibodyPlant = self.plant

        # Add the manipuland to the plant
        temp_idcs = Parser(plant=self.plant).AddModels(
            self.path_to_object,
        )
        assert len(temp_idcs) == 1, f"Only one model should be added; received {len(temp_idcs)}"
        self.manipuland_index = temp_idcs[0]
        self.manipuland_name = self.plant.GetModelInstanceName(self.manipuland_index)

        if and_weld_to is not None:
            # Weld the first frame in the model to the frame given by and_weld_to
            assert isinstance(and_weld_to, Frame), \
                f"and_weld_to must be a Frame; received {type(and_weld_to)}"
            
            manipuland_body_idcs = plant.GetBodyIndices(self.manipuland_index)
            assert len(manipuland_body_idcs) > 0, \
                f"Expected at least one body in the manipuland; received {len(manipuland_body_idcs)}"
            manipuland_body = plant.get_body(manipuland_body_idcs[0])
            frame0 = manipuland_body.body_frame()

            plant.WeldFrames(
                self.plant.world_frame(),
                frame0,
            )

    def connect_to_meshcat(self):
        """
        Description
        -----------
        This method will connect the plant to the meshcat.

        Assumptions
        -----------
        - meshcat_port_number is a positive integer
        """
        # Setup

        # Create meshcat object
        self.meshcat = Meshcat(port=self.meshcat_port_number)  # Object provides an interface to Meshcat
        m_visualizer = MeshcatVisualizer(
            self.meshcat,
        )
        params = MeshcatVisualizerParams(
            role=DrakeRole.kIllustration,
        )
        if self.show_collision_geometries:
            params = MeshcatVisualizerParams(
                role=DrakeRole.kProximity,
            )

        m_visualizer.AddToBuilder(
            self.builder, self.scene_graph, self.meshcat,
            params=params,
        )


from dataclasses import dataclass
import numpy as np
from pydrake.all import (
    Context,
    ModelInstanceIndex,
    MultibodyPlant,
    RigidTransform,
    SpatialVelocity
)

@dataclass
class InitialCondition:
    model_instance_index: ModelInstanceIndex
    pose_wrt_parent: RigidTransform = None # Is parent always world?
    configuration: np.ndarray = None
    target_body_index: int = 0

    def set_initial_configuration(
        self,
        plant: MultibodyPlant,
        diagram_context: Context = None
    ):
        """
        *Description*
        
        Set the initial configuration for the given model instance in the plant.

        Note: Defining the default positions will fail, if the context for the plant
        has already been created. In that case, we also set the positions in the provided
        diagram context (if any).
        """
        # Input Processing
        # - If no configuration is provided, then skip this function
        if self.configuration is None:
            return

        # - If the plant is not yet finalized, then raise an exception!
        if not plant.is_finalized():
            raise RuntimeError("Plant is not yet finalized! Finalize before calling set_initial_configuration()!")

        # - If the configuration value exists, then verify that it has the correct size
        if len(self.configuration) != plant.num_positions(self.model_instance_index):
            raise ValueError(f"Configuration size {len(self.configuration)} does not match plant's num_positions {plant.num_positions(self.model_instance_index)}")

        # Set initial configuration
        plant.SetDefaultPositions(self.model_instance_index, self.configuration)
        print(f"Set default positions for model instance {self.model_instance_index} to {self.configuration}")
        if diagram_context is not None:
            plant.SetPositions(
                plant.GetMyMutableContextFromRoot(diagram_context),
                self.model_instance_index,
                self.configuration
            )

    def set_initial_pose(self, plant: MultibodyPlant, diagram_context: Context = None):
        """
        Description
        -----------
        Set the initial pose for the given model instance in the plant.
        """
        # Input Processing
        # - If no pose is provided, then skip this function
        if not self.pose_wrt_parent:
            return

        # Get the body frame
        attached_bodies = plant.GetBodyIndices(self.model_instance_index)
        target_body_index = attached_bodies[self.target_body_index]

        # Set initial pose
        plant.SetDefaultFreeBodyPose(
            plant.get_body(target_body_index),
            self.pose_wrt_parent)
        
        if diagram_context is not None:
            plant.SetFreeBodyPose(
                plant.GetMyMutableContextFromRoot(diagram_context),
                plant.get_body(target_body_index),
                self.pose_wrt_parent)
            
            plant.SetFreeBodySpatialVelocity(
                body=plant.get_body(target_body_index),
                V_PB=SpatialVelocity.Zero(),
                context=plant.GetMyMutableContextFromRoot(diagram_context),
            )
        


from brom_drake.productions.motion_planning.offline.kinematic import (
    ShelfPlanning1,
    ChemLab1,
)
from brom_drake.productions.motion_planning.offline.dynamic import (
    ChemLab2,
)

__all__ = [
    'ShelfPlanning1',
    'ChemLab1',
]


from brom_drake.control.grippers.gripper_controller import GripperController
from brom_drake.control.grippers.gripper_target import GripperTarget

__all__ = [
    "GripperController",
    "GripperTarget",
]


from enum import Enum, unique

@unique
class ProductionID(Enum):
    """
    **Description**

    A unique ID assigned to each production type.
    """

    #: Undefined Production ID (Used to Detect Errors in Production Construction)
    kNotDefined = "NotDefined" 

    #: Show Me This Model Production
    kShowMeThisModel = "ShowMeThisModel" 

    #: Shelf Planning 1 Production, a simple kinematic motion planning task
    kShelfPlanning1 = "ShelfPlanning1" 
    
    #: Chem Lab Planning 1 Production, a simple kinematic motion planning task
    kChemLab1 = "ChemLab-Planning1" 

    #: Chem Lab Planning 2 Production, a more complex dynamic motion planning task
    kChemLab2 = "ChemLab-Planning2" 
    
    """
    Chem Lab 3, a Pick and Place task in the Chem Lab environment

    .. warning::
        
        This production is currently not functional due to issues with the Chem Lab environment model
    """
    kChemLab3 = "ChemLab-Planning3"
    
    #: Demonstrate Static Grasp, visualizes a static scene of the gripper in a grasping pose with an object
    #: Useful for debugging grasp poses
    kDemonstrateStaticGrasp = "Demonstrate-StaticGrasp" 

    #: Attempt Grasp with Puppeteer, uses the Puppeteer system to attempt a grasp
    #: on a free object that is lying on a movable floor
    kAttemptGraspWithPuppeteer = "Attempt-Grasp-WithPuppeteer"

    #: Attempt Grasp with Static Wrist, attempts a grasp with a static wrist pose
    #: on a free object that is lying on a movable floor
    kAttemptGraspWithStaticWrist = "Attempt-Grasp-WithStaticWrist"


import numpy as np
from pydrake.all import Context
from pydrake.common.value import AbstractValue
from pydrake.multibody.plant import MultibodyPlant
from pydrake.multibody.tree import ModelInstanceIndex
from pydrake.systems.framework import BasicVector

# Internal imports
from brom_drake.control.arms.base_arm_controller import BaseArmController
from brom_drake.control.arms.joint_target import JointTarget


class JointArmController(BaseArmController):
    """
    *Description*

    A controller which imitates the joint control mode of the Kinova gen3 arm. 
    This is designed to work with any arm.

    The controller is an extension of the pydrake.systems.framework.LeafSystem.
    As such, it has input and output ports which are described in detail in the following
    block diagram ::

                                    ------------------------
                                    |                      |
                                    |                      |
        joint_target -------------> |  JointArmController  | ----> applied_arm_torque
        joint_target_type --------> |                      |
                                    |                      |
                                    |                      | ----> measured_ee_pose
        arm_joint_position -------> |                      | ----> measured_ee_twist
        arm_joint_velocity ------>  |                      |
                                    |                      |
                                    |                      |
                                    ------------------------

    The type of target is determined by ee_target_type, and the options are defined in the
    end effector target.
    """
    def __init__(
        self,
        plant: MultibodyPlant,
        arm_model: ModelInstanceIndex,
        end_effector_frame_name: str = "end_effector_frame",
        joint_target_type: JointTarget = JointTarget.kPosition,
    ):
        BaseArmController.__init__(
            self, plant, arm_model,
            end_effector_frame_name=end_effector_frame_name,
        )

        # Define input and Output Ports
        self.joint_target_port, self.joint_target_type_port = None, None
        self.define_joint_target_input_ports(joint_target_type)

        # Define output ports
        self.DeclareVectorOutputPort(
            "applied_arm_torque",
            BasicVector(self.plant.num_actuators()),
            self.CalcArmTorques
        )

    def define_joint_target_input_ports(self, joint_target_type: JointTarget):
        """
        *Description*
        
        Define the input ports for the JointArmController system
        
        *Parameters*

        joint_target_type: JointTarget
            The type of joint target being provided to the controller.
        """
        self.joint_target_port = self.DeclareVectorInputPort(
            "joint_target",
            BasicVector(self.plant.num_positions(self.arm)),
        )
        self.joint_target_type_port = self.DeclareAbstractInputPort(
            "joint_target_type",
            AbstractValue.Make(joint_target_type),
        )

    def CalcArmTorques(self, context: Context, output: BasicVector):
        """
        *Description*
        
        Calculate the arm torques based on the joint target and the current state of the arm.
        
        *Parameters*

        context: pydrake.systems.framework.Context
            The context of the system at the current timestep.

        output: pydrake.systems.framework.BasicVector
            The output vector where the calculated torques will be stored.
        """
        # Get the current state of the arm
        q = self.arm_joint_position_port.Eval(context)
        qd = self.arm_joint_velocity_port.Eval(context)

        # Set our "internal plant's state" to the current state of the arm
        self.plant.SetPositions(self.context, self.arm, q)
        self.plant.SetVelocities(self.context, self.arm, qd)

        # Calculate the torque needed to compensate for gravity
        tau_g = -self.plant.CalcGravityGeneralizedForces(self.context)

        # Indicate what type of command we're recieving
        target_type = self.joint_target_type_port.Eval(context)

        if target_type == JointTarget.kPosition:
            q_des = self.joint_target_port.Eval(context)
            q_err = q_des - q
            qd_des = 0 * q_err
            kp = 3e3
            kd = 2*np.sqrt(3e3)
            tau = kp * q_err + kd * (qd_des - qd) + tau_g

        else:
            raise NotImplementedError(
                f"Only position joint control is currently supported; received {target_type}"
            )

        output.SetFromVector(tau)


from dataclasses import dataclass
import numpy as np
from pydrake.all import (
    RigidTransform,
)

@dataclass
class ProximityPosePlanDispenserConfig:
    proximity_limit: float = 1e-1   # The proximity limit for the poses.
    translation_weight: float = 1.0 # The weight for the translation distance in the distance calculation.
    orientation_weight: float = 0.0 # The weight for the orientation distance in the distance calculation.

    def distance_between_poses(
        self,
        pose1: RigidTransform,
        pose2: RigidTransform,
    ) -> float:
        """
        Description
        -----------
        This method computes the distance between two poses.

        Arguments
        ---------
        pose1 : RigidTransform
            The first pose.
        pose2 : RigidTransform
            The second pose.
        """
        # Setup
        Q_translation = self.translation_weight
        Q_orientation = self.orientation_weight
        # 
        translation_distance = np.linalg.norm(pose1.translation() - pose2.translation())
        orientation_distance = np.linalg.norm(pose1.rotation().matrix() - pose2.rotation().matrix())
        return Q_translation * translation_distance + Q_orientation * orientation_distance

    def in_proximity(
        self,
        pose1: RigidTransform,
        pose2: RigidTransform,
    ) -> bool:
        """
        Description
        -----------
        This method checks if two poses are within the proximity of each other.

        Arguments
        ---------
        pose1 : RigidTransform
            The first pose.
        pose2 : RigidTransform
            The second pose.
        """
        # Setup
        return self.distance_between_poses(pose1, pose2) < self.proximity_limit


from brom_drake.utils.angles import RollPitchYawAngle, rpy_equivalent_body_rotation_order
from dataclasses import dataclass
import numpy as np
from pydrake.all import (
    Adder,
    ConstantVectorSource,
    Demultiplexer,
    DiagramBuilder,
    Frame,
    LeafSystem,
    Multiplexer,
    Parser,
    PassThrough,
    PidController,
    PrismaticJoint,
    RevoluteJoint,
    Joint,
    JointActuator,
)
from pydrake.multibody.plant import MultibodyPlant
from pydrake.multibody.tree import ModelInstanceIndex
from typing import List, Tuple

from brom_drake.utils.leaf_systems.named_vector_selection_system import define_named_vector_selection_system
from brom_drake.utils.model_instances import get_all_bodies_in

# Internal Imports
from .configuration import Configuration as PuppetmakerConfiguration
from .puppet_signature import PuppetSignature, PuppeteerJointSignature, AllJointSignatures
from brom_drake.file_manipulation.urdf import SimpleShapeURDFDefinition
from brom_drake.file_manipulation.urdf.shapes import SphereDefinition
from brom_drake.utils.leaf_systems.rigid_transform_to_vector_system import (
    RigidTransformToVectorSystem,
    RigidTransformToVectorSystemConfiguration,
)

class Puppetmaker:
    """
    *Description*
    
    This class turns a "free body" in Drake into a "puppet" whose pose and orientation
    can be controlled using 6 actuated joints (3 prismatic and 3 revolute).

    This object is used to:
    - Add "invisible" actuators for moving a free body (i.e., `add_strings_for` or `add_actuators_for`), and
    - Control actuate a free body that has been added with the expected "puppet" actuators (i.e., `add_puppet_controller_for)

    .. important::
        The 2 functions of the puppetmaker are designed to be used at different times.
        The `add_strings_for` or `add_actuators_for` methods MUST be called BEFORE finalizing the plant.
        The `add_puppet_controller_for` method MUST be called AFTER finalizing the plant.
    
    *Notes*
    
    - The puppet is assumed to be a free body (i.e., it has no existing joints connecting it to the world).
      After adding the puppet actuators, the puppet will NO LONGER be a free body.
    - You must call add_strings_for or add_actuators_for BEFORE finalizing the plant.
    """
    def __init__(
        self,
        plant: MultibodyPlant,
        puppet_anchored_to: Frame = None,
        frame_on_puppet: Frame = None,
        name: str = None,
        sphere_radius: float = None,
        config: PuppetmakerConfiguration = None,
    ):
        # Save some helpful internal variables
        self.plant = plant

        # Create config
        self.config = self.config_from_initialization_params(
            puppet_anchored_to=puppet_anchored_to,
            frame_on_puppet=frame_on_puppet,
            name=name,
            sphere_radius=sphere_radius,
            config=config,
        )

    def add_actuated_prismatic_joint(
        self,
        axis_dimension: int,
        previous_frame: Frame,
        next_frame: Frame,
        joint_name: str,
    ) -> Tuple[PrismaticJoint, JointActuator]:
        """
        **Description**
        
        This method will add an prismatic joint in the direction `axis_dimension`
        between `previous_frame` and `next_frame`, along with an actuator to control it.
        
        **Parameters**
        
        axis_dimension: int
            The axis to translate in (0 for X, 1 for Y, 2 for Z).

        previous_frame: Frame
            The frame of the parent side of the joint.

        next_frame: Frame
            The frame of the child side of the joint.

        joint_name: str
            The name of the joint we are creating.

        **Returns**
        
        translation_joint_ii: PrismaticJoint
            The created prismatic joint.

        translation_actuator_ii: JointActuator
            The created actuator for the prismatic joint.        
        """
        # Setup
        plant: MultibodyPlant = self.plant

        # Create the direction to translate in
        axis_ii = [0, 0, 0]
        axis_ii[axis_dimension] = 1.0

        # Create Joint
        translation_joint_ii = plant.AddJoint(
            PrismaticJoint(
                name=joint_name,
                frame_on_parent=previous_frame,
                frame_on_child=next_frame,
                axis=axis_ii,
            )
        )

        # Create actuator for joint
        translation_actuator_ii = plant.AddJointActuator(
            joint_name + "_actuator",
            translation_joint_ii,
        )
        return translation_joint_ii, translation_actuator_ii

    def add_actuated_revolute_joint(
        self,
        angle_type: RollPitchYawAngle,
        previous_frame: Frame,
        next_frame: Frame,
        joint_name: str,
    ) -> Tuple[Joint, JointActuator]:
        """
        *Description*

        This method will add a revolute joint in the direction `axis_dimension`
        between `previous_frame` and `next_frame`, along with an actuator to control it.

        *Parameters*

        axis_dimension: int
            The axis to rotate about 
        """
        # Setup
        plant: MultibodyPlant = self.plant

        # Create the direction to rotate in
        axis_ii = [0, 0, 0]
        match angle_type:
            case RollPitchYawAngle.kRoll:
                axis_ii[0] = 1.0
            case RollPitchYawAngle.kPitch:
                axis_ii[1] = 1.0
            case RollPitchYawAngle.kYaw:
                axis_ii[2] = 1.0
            case _:
                raise ValueError(
                    f"Invalid angle_type '{angle_type}'; must be one of {list(RollPitchYawAngle)}"
                )

        # Create Joint
        rotation_joint_ii = plant.AddJoint(
            RevoluteJoint(
                name=joint_name,
                frame_on_parent=previous_frame,
                frame_on_child=next_frame,
                axis=axis_ii,
            )
        )

        # Create actuator for joint
        rotation_actuator_ii = plant.AddJointActuator(
            joint_name + "_actuator",
            rotation_joint_ii,
        )
        
        # Return
        return rotation_joint_ii, rotation_actuator_ii

    def add_all_actuated_prismatic_joints(
        self,
        target_model: ModelInstanceIndex,
        massless_sphere_indices: List[ModelInstanceIndex],
    ) -> List[PuppeteerJointSignature]:
        # Setup
        config = self.config
        plant: MultibodyPlant = self.plant

        # Prepare for loop
        joints = []
        joint_actuators = []

        # Create Three Translation Joints
        previous_frame = config.frame_on_parent
        translation_joint_names = self.translation_joint_names(target_model)
        for axis_dimension, joint_name_ii in enumerate(translation_joint_names):
            # Compute the next frame to connect to from sphere_ii
            # It will be the only frame in the sphere model
            sphere_ii = massless_sphere_indices[axis_dimension]
            sphere_ii_bodies = plant.GetBodyIndices(sphere_ii)
            sphere_ii_frame = plant.get_body(sphere_ii_bodies[0]).body_frame()

            joint_ii, joint_actuator_ii =self.add_actuated_prismatic_joint(
                axis_dimension=axis_dimension,
                previous_frame=previous_frame,
                next_frame=sphere_ii_frame,
                joint_name=joint_name_ii,
            )

            # Update previous frame for next iteration of loop
            previous_frame = sphere_ii_frame

            # Save new joints and actuators
            joints.append(joint_ii)
            joint_actuators.append(joint_actuator_ii)

        return [
            PuppeteerJointSignature(
                joint=joints[ii],
                joint_actuator=joint_actuators[ii]
            )
            for ii in range(len(joints))
        ]

    def add_all_actuated_revolute_joints(
        self,
        target_model: ModelInstanceIndex,
        massless_sphere_indices: List[ModelInstanceIndex],
    ) -> List[PuppeteerJointSignature]:
        """
        *Description*

        This function creates the 3 actuated revolute joints needed to
        control the orientation of the puppet.

        Because the puppet's orientation is controlled using roll-pitch-yaw angles,
        the order of the revolute joints will correspond to the standard body rotation
        order for roll-pitch-yaw (note this is not roll first, then pitch, then yaw).
        

        *Notes*

        This function expects 3 massless sphere inputs.
        The first sphere in the list should be the one representing the last 
        translation sphere (i.e., the one we should connect the first revolute sphere to.)
        """
        # Setup
        config = self.config
        plant: MultibodyPlant = self.plant

        # Prepare output containers
        joints = []
        joint_actuators = []

        # Find the previous joint to connect the FIRST rotation joint to
        # (This should be the sphere used for the last translation joint.)
        sphere2 = massless_sphere_indices[0]
        sphere2_bodies = plant.GetBodyIndices(sphere2)
        previous_frame = plant.get_body(sphere2_bodies[0]).body_frame()

        # Create The First 2 Rotation Joints
        all_rpy_angles = rpy_equivalent_body_rotation_order()
        for axis_dimension, rpy_angle in enumerate(all_rpy_angles[:2]):
            joint_name_ii = self.rotation_joint_name(
                target_model=target_model,
                rpy_selection=rpy_angle,
            )
            
            # Compute the next frame to connect to from sphere_ii
            # It will be the only frame in the sphere model
            sphere_ii = massless_sphere_indices[axis_dimension+1]
            sphere_ii_bodies = plant.GetBodyIndices(sphere_ii)
            sphere_ii_frame = plant.get_body(sphere_ii_bodies[0]).body_frame()

            joint_ii, joint_actuator_ii = self.add_actuated_revolute_joint(
                angle_type=rpy_angle,
                previous_frame=previous_frame,
                next_frame=sphere_ii_frame,
                joint_name=joint_name_ii,
            )

            # Update previous frame for next iteration of loop
            previous_frame = sphere_ii_frame

            # Save new joints and actuators
            joints.append(joint_ii)
            joint_actuators.append(joint_actuator_ii)

        # Connect the last rotation joint to the puppet
        frame_on_puppet = self.find_frame_on_puppet(target_model)
        final_joint_name = self.rotation_joint_name(
            target_model=target_model,
            rpy_selection=all_rpy_angles[2],
        )
        joint_ii, joint_actuator_ii = self.add_actuated_revolute_joint(
            angle_type=all_rpy_angles[2],
            previous_frame=previous_frame,
            next_frame=frame_on_puppet,
            joint_name=final_joint_name,
        )

        # Save new joints and actuators
        joints.append(joint_ii)
        joint_actuators.append(joint_actuator_ii)

        return [
            PuppeteerJointSignature(
                joint=joints[ii],
                joint_actuator=joint_actuators[ii]
            )
            for ii in range(len(joints))
        ]

    def add_strings_for(
        self,
        target_model: ModelInstanceIndex,
    ) -> PuppetSignature:
        """
        *Description*

        Adds the necessary actuators for the puppet's joints.
        
        *Parameters*
        
        target_model: pydrake.multibody.tree.ModelInstanceIndex
            The model instance index of the puppet to add actuators for.
        """
        return self.add_actuators_for(target_model)

    def add_actuators_for(
        self,
        target_model: ModelInstanceIndex,
    ) -> PuppetSignature:
        """
        *Description*
        
        Creates actuated joints on the object we want to be the "puppet" so that we can control
        its entire pose. This will be 6 joints:

        1. Translation in X
        2. Translation In Y
        3. Translation In Z
        4. Rotation (Roll)
        5. Rotation (Pitch)
        6. Rotation (Yaw)

        *Parameters*

        target_model: pydrake.multibody.tree.ModelInstanceIndex
            The model instance index of the puppet to add actuators for.
        """
        # Setup
        config: PuppetmakerConfiguration = self.config
        plant: MultibodyPlant = self.plant

        # Input Checking
        if plant.is_finalized():
            raise ValueError(
                f"Plant should not be finalized before calling `create_actuators_for_puppet()` method."
            )

        # Create some fictitious bodies which will be used to connect some new actuators
        translation_ghost_models, revolute_ghost_models = self.create_ghost_bodies_for_actuators()

        # Create Three Translation Joints
        prismatic_signatures = self.add_all_actuated_prismatic_joints(
            target_model=target_model,
            massless_sphere_indices=translation_ghost_models,
        )

        # Create the rotational joints
        revolute_signatures = self.add_all_actuated_revolute_joints(
            target_model=target_model,
            massless_sphere_indices=[translation_ghost_models[-1]] + revolute_ghost_models,
        )

        # Return
        return PuppetSignature(
            name=f"[{config.name}] Puppet Signature for {plant.GetModelInstanceName(target_model)}",
            model_instance_index=target_model,
            prismatic_ghost_bodies=translation_ghost_models,
            revolute_ghost_bodies=revolute_ghost_models,
            joints=AllJointSignatures(
                prismatic=prismatic_signatures,
                revolute=revolute_signatures,
            ),
        )

    def add_feedforward_sum_system_for(
        self,
        signature: PuppetSignature,
        builder: DiagramBuilder,
    ) -> Tuple[Adder, ConstantVectorSource]:
        """
        Description
        -----------
        This method will create a simple Adder system to add feedforward
        inputs to the PID controller output.
        """
        # Setup
        plant: MultibodyPlant = self.plant
        n_actuators_for_puppet = signature.n_joints

        # Create the adder system
        feedforward_adder = builder.AddNamedSystem(
            name=f"[{self.config.name}] Feedforward Adder for puppeteering \"{signature.name}\"",
            system=Adder(
                num_inputs=2,
                size=n_actuators_for_puppet,
            ),
        )

        # Create a constant vector source to provide feedforward gravity compensation
        # for the linear actuators
        gravity_field = plant.gravity_field()
        gravity_vector = gravity_field.gravity_vector() # Returns the gravity vector in world frame in units of m/s^2

        # Calculate mass of the puppet
        bodies_in_puppet = get_all_bodies_in(plant, signature.model_instance_index)
        total_mass = np.sum(
            [body_ii.default_mass() for body_ii in bodies_in_puppet]
        )

        gravity_compensation = np.zeros((n_actuators_for_puppet,))
        gravity_compensation[:3] = -total_mass * gravity_vector

        # Create the constant vector source
        gravity_compensation_source = builder.AddNamedSystem(
            name=f"[{self.config.name}] Gravity Compensation for puppeteering \"{signature.name}\"",
            system=ConstantVectorSource(gravity_compensation),
        )

        # Connect the gravity compensation to the adder
        builder.Connect(
            gravity_compensation_source.get_output_port(0),
            feedforward_adder.get_input_port(0),
        )

        return feedforward_adder, gravity_compensation_source

    def add_puppet_controller_for(
        self,
        signature: PuppetSignature,
        builder: DiagramBuilder,
        Kp: np.ndarray = None,
        Kd: np.ndarray = None,
    ) -> Tuple[RigidTransformToVectorSystem, PassThrough|None]:
        """
        *Description*
        
        This method will create a simple PID controller to actuate the puppet
        to a desired pose.

        *Parameters*
        
        signature : PuppetSignature
            The signature of the puppet to be controlled.

        builder : DiagramBuilder
            The diagram builder to add the controller to.

        Kp : np.ndarray, optional
            The proportional gain for the PID controller. If not provided, a default value will be used.
            
        Kd : np.ndarray, optional
            The derivative gain for the PID controller. If not provided, a default value will be used.

        *Returns*
        
        pose_to_vector_system : RigidTransformToVectorSystem
            The system that converts the desired pose to a vector format.
            This system is what provides the target to the PID controller.
            You can think of this as the "reference" signal for the controller.

        passthrough_system : PassThrough|None
            A passthrough system that can be used to provide inputs to the
            "puppet" if the Puppet contains actuated joints.
            If the puppet does not contain any actuated joints, then this will be None.
        """
        # Setup
        plant: MultibodyPlant = self.plant
        n_actuators_for_puppet = signature.n_joints

        # Input Processing
        # - Verify that plant is finalized already
        if not self.plant.is_finalized():
            raise ValueError("Plant is not finalized yet.\nPlant must be finalized before calling this method!")

        if Kp is None:
            bodies_in_puppet = plant.GetBodyIndices(signature.model_instance_index)
            m, gravity = 0.0, 9.81
            for body_ii in bodies_in_puppet:
                m += plant.get_body(body_ii).default_mass()
            Kp = np.zeros((n_actuators_for_puppet,))
            Kp[:3] = np.array([1e2*m*gravity]*3)
            Kp[3:] = np.array([1e0*m*gravity]*3)

        if Kd is None:
            Kd = np.sqrt(Kp)

        # Create a demultiplexer to combine all actuator inputs
        actuator_demux, pass_through_system = self.create_actuator_demux(signature, builder)

        # Create the PID Controller
        controller: PidController = builder.AddSystem(
            PidController(
                kp=Kp,
                ki=np.zeros(Kp.shape),
                kd=Kd,
            )
        )

        # Create the feedforward adder system
        feedforward_adder, _ = self.add_feedforward_sum_system_for(signature, builder)

        # Connect the feedforward adder to the controller
        self.combine_feedforward_and_controller_for(
            builder=builder,
            feedforward_adder=feedforward_adder,
            controller=controller,
        )

        # Connect the adder (which combines the PID Controller and the feedforward term)
        # to the demultiplexer
        builder.Connect(
            feedforward_adder.get_output_port(0),
            actuator_demux.get_input_port(0),
        )

        # Connect a PoseToVector System to the PID controller
        # as reference.
        p2v_config = RigidTransformToVectorSystemConfiguration(
            name=f"[{self.config.name}]Pose to vector for puppeteering \"{signature.name}\"",
            output_format="vector_xyz_euler(rpy)",
        )
        pose_to_vector = builder.AddSystem(
            RigidTransformToVectorSystem(p2v_config)
        )

        # Reference signal is a state which contains two pieces: a pose and a twist
        # - Target Twist is zero
        zero_twist_source = builder.AddSystem(
            ConstantVectorSource(np.zeros((6,)))
        )

        target_state_mux = builder.AddNamedSystem(
            name=f"[{self.config.name}] Target State Mux for puppeteering \"{signature.name}\"",
            system=Multiplexer(input_sizes=[6, 6]),
        )

        # Connect target state mux inputs
        # - Pose from pose_to_vector (transformed so that the order of its
        #   output matches the expected input of the controller; controller
        #   expects [x, y, z, yaw, pitch, roll])

        rpy_ordering_selection_system = define_named_vector_selection_system(
            all_element_names=["x", "y", "z", "roll", "pitch", "yaw"],
            sequence_of_names_for_output=["x", "y", "z", "yaw", "pitch", "roll"],
        )
        builder.AddNamedSystem(
            name=f"[{self.config.name}] RPY Ordering Selection for puppeteering \"{signature.name}\"",
            system=rpy_ordering_selection_system,
        )

        builder.Connect(
            pose_to_vector.get_output_port(0),
            rpy_ordering_selection_system.get_input_port(0),
        )
        builder.Connect(
            rpy_ordering_selection_system.get_output_port(0),
            target_state_mux.get_input_port(0),
        )
        # - Twist from zero_twist_source
        builder.Connect(
            zero_twist_source.get_output_port(0),
            target_state_mux.get_input_port(1),
        )
        
        # Connect mux output (i.e., the desired state) to PID
        builder.Connect(
            target_state_mux.get_output_port(0),
            controller.get_input_port_desired_state(),
        )

        self.connect_plant_to_controller(builder, controller, signature)

        return pose_to_vector, pass_through_system

    def combine_feedforward_and_controller_for(
        self,
        builder: DiagramBuilder,
        feedforward_adder: Adder,
        controller: PidController,
    ) -> Tuple[Adder, ConstantVectorSource]:
        """
        Description
        -----------
        This method will connect the output of the PID controller to one input
        of the feedforward adder.
        """
        # Setup

        # Connect the output of the controller to one input of the adder
        builder.Connect(
            controller.get_output_port(0),
            feedforward_adder.get_input_port(1),
        )

        return feedforward_adder, None

    def config_from_initialization_params(
        self,
        puppet_anchored_to: Frame|None,
        frame_on_puppet: Frame|None,
        name: str|None,
        sphere_radius: float = None,
        config: PuppetmakerConfiguration = None,
    ) -> PuppetmakerConfiguration:
        # Input Checking
        if config is not None:
            return config # If the config was provided, then simply return that.
        
        # If config is not given, then we will try to create it.

        # Setup
        plant: MultibodyPlant = self.plant

        # Choose Frame to Anchor things to
        if puppet_anchored_to is None:
            puppet_anchored_to = plant.world_frame()

        # Choose name of the puppeteer
        if name is None:
            name = "Puppeteer"

        # Create config with these initial values
        temp_config = PuppetmakerConfiguration(
            frame_on_parent=puppet_anchored_to,
            name=name,
            frame_on_child=frame_on_puppet,
        )

        # Update config with other optional values
        if sphere_radius is not None:
            temp_config.sphere_radius = sphere_radius

        return temp_config

    def connect_plant_to_controller(
        self,
        builder: DiagramBuilder,
        controller: PidController,
        signature: PuppetSignature,
    ) -> None:
        """
        *Description*

        Connect the plant to the controller.


        """
        # Get some of the internal variables from the puppetmaker
        plant: MultibodyPlant = self.plant

        # Create a system that will share the state of the puppet with the controller
        # This system will be a multiplexer that combines the position and velocity
        # states of each of the models in the puppet.

        # - Create a massive multiplexer to combine all of the states (2 each) of the models (there should be 6) in the puppet
        state_mux = builder.AddNamedSystem(
            name=f"[{self.config.name}] State Mux for puppeteering \"{signature.name}\"",
            system=Multiplexer(num_scalar_inputs=2*signature.n_models),
        )

        # - Create connections from the state of each puppet model (comes from the plant)
        #   to new demultiplexers
        all_state_demuxes: List[Demultiplexer] = []
        for ii, model_ii in enumerate(signature.all_models):
            # Create demultiplexer for this model
            n_state_ii = plant.num_multibody_states(model_ii)
            print(f"Model #{ii}'s states: {plant.GetStateNames(model_ii)}")
            state_demux_ii: Demultiplexer = builder.AddSystem(
                Demultiplexer(size=n_state_ii),
            )

            # Connect the plant state to the demux
            builder.Connect(
                plant.get_state_output_port(model_ii),
                state_demux_ii.get_input_port(0),
            )

            all_state_demuxes.append(state_demux_ii)

        # - Connect all demuxes to the overall state mux
        for ii, demux_ii in enumerate(all_state_demuxes):
            model_ii = signature.all_models[ii]

            # Find position component of puppet joint state and
            # send it to the state_mux
            position_ii_state_index = self.find_index_of_position_state_in(
                model_instance_index=model_ii,
                signature=signature,
            )
            builder.Connect(
                demux_ii.get_output_port(position_ii_state_index),
                state_mux.get_input_port(ii),
            )

            # Find velocity component of puppet joint state and
            # send it to the state_mux
            velocity_ii_state_index = self.find_index_of_velocity_state_in(
                model_instance_index=model_ii,
                signature=signature,
            )
            builder.Connect(
                demux_ii.get_output_port(velocity_ii_state_index),
                state_mux.get_input_port(signature.n_models + ii),
            )

        # Connect the output of the state demux to the controller
        builder.Connect(
            state_mux.get_output_port(0),
            controller.get_input_port_estimated_state(),
        )

    def create_actuator_demux(
        self,
        signature: PuppetSignature,
        builder: DiagramBuilder,
    ) -> Tuple[Demultiplexer, PassThrough|None]:
        """
        *Description*
        
        This method creates a demultiplexer to split the actuator inputs
        to the puppet's various models.
        """
        # Setup
        plant: MultibodyPlant = self.plant
        n_actuators_for_puppet = signature.n_joints

        # Create a demultiplexer to combine all actuator inputs
        actuator_demux = builder.AddSystem(
            Demultiplexer(size=n_actuators_for_puppet),
        )

        # Connect demultiplexer to each of the actuators
        pass_through_system = None # Default output, if the puppet has no actuators

        for ii, model_ii in enumerate(signature.all_models):
            if plant.num_actuators(model_ii) == 0:
                continue

            print(f"n_actuators for model {plant.GetModelInstanceName(model_ii)}: {plant.num_actuators(model_ii)}")
            # print(f"Model #{ii}'s actuators: {plant.GetActuatorNames(model_ii)}")

            # for port_jj in plant.GetActuatorNames(model_ii):
            #     print(f"Actuator on model {plant.GetModelInstanceName(model_ii)}: {port_jj}")
                
            if plant.num_actuators(model_ii) == 1:
                # If there is only one actuator for this model,
                # then it is the actuator that we created and 
                # we can connect it directly to the demux input

                builder.Connect(
                    actuator_demux.get_output_port(ii),
                    plant.get_actuation_input_port(model_ii),
                )
            else:
                # If there is more than one actuator for this model,
                # then we will need to create a passthrough system to
                # pass the "default" actuator inputs for the puppet's (default) actuated joints
                # and a separate passthrough system to pass the "puppet" actuator input for puppeteering.
                
                default_actuator_pass_through, puppeteer_pass_through = self.create_actuator_passthrough_for_puppet_with_actuated_joints(
                    signature=signature,
                    builder=builder,
                )
                pass_through_system = default_actuator_pass_through

                # Connect the puppeteer passthrough to the demux
                builder.Connect(
                    actuator_demux.get_output_port(ii),
                    puppeteer_pass_through.get_input_port(),
                )                

        return actuator_demux, pass_through_system
    
    def create_actuator_passthrough_for_puppet_with_actuated_joints(
        self,
        signature: PuppetSignature,
        builder: DiagramBuilder,
    ) -> Tuple[PassThrough, PassThrough]:
        """
        *Description*
        
        This method creates two passthrough systems:
        1. passes the "default" actuator inputs for the puppet's (default) actuated joints,
        2. passes the "puppet" actuator input for puppeteering.

        *Returns*

        default_actuator_inputs_passthrough : PassThrough
            A passthrough system that passes the default actuator inputs for the puppet's (default)
            actuated joints.

        puppet_actuator_input : PassThrough
            A passthrough system that passes the puppet actuator input for puppeteering.
        """
        # Setup
        plant: MultibodyPlant = self.plant
        puppet_model_idx = signature.model_instance_index
        n_default_actuators_in_puppet = plant.num_actuators(puppet_model_idx)
        
        # Create a Multiplexer system to combine all actuator inputs
        passthrough_mux = builder.AddNamedSystem(
            name=f"[{self.config.name}] Total Actuation Mux for puppeteering \"{signature.name}\" WITH default inputs needed",
            system=Multiplexer(input_sizes=[n_default_actuators_in_puppet-1,1]),
        )

        # Create a passthrough system to pass the combined actuator inputs to the puppet
        default_actuator_inputs_passthrough = builder.AddSystem(
            PassThrough(value=np.zeros((n_default_actuators_in_puppet-1,))),
        )

        # Connect the default actuator inputs to the multiplexer
        builder.Connect(
            default_actuator_inputs_passthrough.get_output_port(0),
            passthrough_mux.get_input_port(0),
        )

        # Connect the puppet actuator input to the multiplexer
        puppet_actuator_input = builder.AddSystem(
            PassThrough(value=np.zeros((1,))),
        )
        builder.Connect(
            puppet_actuator_input.get_output_port(0),
            passthrough_mux.get_input_port(1),
        )

        # connect the multiplexer to the plant's input for the
        # puppet's FULL set of actuators (both the ones we added and the ones that it already contains)
        builder.Connect(
            passthrough_mux.get_output_port(),
            plant.get_actuation_input_port(puppet_model_idx),
        )

        return default_actuator_inputs_passthrough, puppet_actuator_input

    def create_ghost_bodies_for_actuators(
        self,
    ) -> Tuple[List[ModelInstanceIndex], List[ModelInstanceIndex]]:
        """
        Description
        -----------
        This method creates a set of simple spheres which we will use to connect
        actuators to the "puppet". We will reuse the simple shape API's in Brom
        to create a set of new URDFs that will have (i) no mass, and
        (ii) no collision geometry.
        """
        # Setup
        config = self.config

        # Create list of ModelInstances
        translation_ghost_models: List[ModelInstanceIndex] = []
        revolute_ghost_models: List[ModelInstanceIndex] = []

        # Create the spheres that connect the translation joints
        for translation_sphere_idx in range(2):
            # Create the sphere name
            sphere_ii_name = f"[{config.name}]translation_{translation_sphere_idx}_to_translation_{translation_sphere_idx+1}"

            # Update the output list
            translation_ghost_models.append(
                self.add_massless_sphere_to_plant_with_name(
                    name=sphere_ii_name,
                    color=config.sphere_color,
                )
            )

        # Create the sphere that connects the last translation joint to the first rotation joint
        sphere_2_name = f"[{config.name}]translation_2_to_rotation_0"
        translation_ghost_models.append(
            self.add_massless_sphere_to_plant_with_name(
                name=sphere_2_name,
                color=config.sphere_color,
            )
        )

        # Create the spheres that connect the rotation joints
        for rotation_sphere_idx in [0, 1]:
            # Create the name of the sphere
            sphere_jj_name = f"[{config.name}]rotation_{rotation_sphere_idx}_to_rotation_{rotation_sphere_idx+1}"
            # Update the output list
            revolute_ghost_models.append(
                self.add_massless_sphere_to_plant_with_name(
                    name=sphere_jj_name,
                    color=config.sphere_color,
                )
            )

        return translation_ghost_models, revolute_ghost_models

    def add_massless_sphere_to_plant_with_name(self, name: str, color: np.ndarray = None) -> ModelInstanceIndex:
        """
        *Description*
        
        This method adds a sphere with:

        - nearly zero mass
        - no collision geometry,
        - and an associated color
        
        to the Puppeteer's plant. 
        
        The sphere's name will be assigned by the input to the function.
        """
        # Setup
        config = self.config
        plant: MultibodyPlant = self.plant

        # Input Processing
        if color is None:
            color = np.array([0.0, 0.0, 1.0, 0.4])

        # Create the sphere shape
        sphere_shape = SphereDefinition(radius=config.sphere_radius)

        # Create a URDF for this sphere
        sphere_urdf_defn = SimpleShapeURDFDefinition(
            name=name,
            shape=sphere_shape,
            create_collision=False,
            mass=config.sphere_mass,
            color=color,
        )
        sphere_urdf = sphere_urdf_defn.write_to_file()

        # Add the sphere to the plant
        sphere_model_index = Parser(plant=plant).AddModels(sphere_urdf)[0]

        return sphere_model_index

    def find_frame_on_puppet(self, target_model: ModelInstanceIndex):
        # Setup
        config = self.config
        plant: MultibodyPlant = self.plant

        # Choose the frame on the puppet to attach actuators to
        if config.frame_on_child is not None:
            return self.config.frame_on_child
        else:
            # Search through all frames of the model
            bodies_in_puppet = plant.GetBodyIndices(target_model)
            first_body = plant.get_body(bodies_in_puppet[0])
            first_frame_on_first_body = first_body.body_frame()
            return first_frame_on_first_body

    def find_index_of_position_state_in(
        self,
        model_instance_index: ModelInstanceIndex,
        signature: PuppetSignature,
    ) -> int|None:
        """
        Description
        -----------
        This method finds the names of the "position" component for the
        state of the joint connected to model_instance_index.
        """
        # Setup
        plant: MultibodyPlant = self.plant
        target_model: ModelInstanceIndex = signature.model_instance_index

        # For each state state name,
        for ii, state_name_ii in enumerate(plant.GetStateNames(model_instance_index)):
            print("state name: ", state_name_ii)
            # Check to see if name is either:
            # - A Translational joint state name (for the target model)
            for translation_joint_name_jj in self.translation_joint_names(target_model, remove_puppeteer_prefix=True):
                print("- translation_joint_name: ", translation_joint_name_jj)
                if (translation_joint_name_jj in state_name_ii) and ("_x" in state_name_ii):
                    return ii
                
            # - A Rotational Joint state name
            for rotational_joint_name_jj in self.rotation_joint_names(target_model, remove_puppeteer_prefix=True):
                print("- rotational_joint_name: ", rotational_joint_name_jj)
                if (rotational_joint_name_jj in state_name_ii) and ("_q" in state_name_ii):
                    return ii
                
        # If none of the states contained a joint name, then return None
        return None


    def find_index_of_velocity_state_in(
        self,
        model_instance_index: ModelInstanceIndex,
        signature: PuppetSignature,
    ) -> int|None:
        """
        *Description*
        
        This method finds the names of the "velocity" component for the
        state of the joint connected to model_instance_index.

        *Parameters*

        model_instance_index: pydrake.multibody.tree.ModelInstanceIndex
            The model instance index of the model to search through.

        signature: PuppetSignature
            The signature of the puppet being controlled.
        """
        # Setup
        plant: MultibodyPlant = self.plant
        target_model: ModelInstanceIndex = signature.model_instance_index

        # For each state state name,
        for ii, state_name_ii in enumerate(plant.GetStateNames(model_instance_index)):
            # Check to see if name is either:
            # - A Translational joint state name
            for translation_joint_name_jj in self.translation_joint_names(target_model, remove_puppeteer_prefix=True):
                if (translation_joint_name_jj in state_name_ii) and ("_v" in state_name_ii):
                    return ii
                
            # - A Rotational Joint state name
            for rotational_joint_name_jj in self.rotation_joint_names(target_model, remove_puppeteer_prefix=True):
                if (rotational_joint_name_jj in state_name_ii) and ("_w" in state_name_ii):
                    return ii
                
        # If none of the states contained a joint name, then return None
        return None

    def rotation_joint_name(
        self,
        target_model: ModelInstanceIndex,
        rpy_selection: RollPitchYawAngle,
        remove_puppeteer_prefix: bool = False,
    ) -> str:
        """
        *Description*
        
        This method returns the name of a specific rotation joint created for the puppet.

        *Parameters*

        target_model: pydrake.multibody.tree.ModelInstanceIndex
            The model instance index of the puppet.

        axis_dimension: int
            The axis dimension (0, 1, or 2) corresponding to the rotation joint to get the name for.

        remove_puppeteer_prefix: bool, optional
            If True, the returned joint name will not include the puppeteer prefix.
            Default is False.
        """
        # Setup
        config = self.config
        plant: MultibodyPlant = self.plant

        # Find name of plant
        target_model_name = plant.GetModelInstanceName(target_model)
        puppeteer_prefix = "" if remove_puppeteer_prefix else f"[{config.name}]"

        return f"{puppeteer_prefix}{target_model_name}_rotation_joint_{str(rpy_selection)}"

    def rotation_joint_names(
        self,
        target_model: ModelInstanceIndex,
        remove_puppeteer_prefix: bool = False,
    ) -> List[str]:
        """
        *Description*
        
        This method returns the names of all rotation joints created for the puppet.

        *Parameters*

        target_model: pydrake.multibody.tree.ModelInstanceIndex
            The model instance index of the puppet.

        remove_puppeteer_prefix: bool, optional
            If True, the returned joint names will not include the puppeteer prefix.
            Default is False.
        """
        # The rotation axis names will be ordered according to the
        # standard body rotation order assigned to RPY (note this is not roll first, then pitch, then yaw)
        return [
            self.rotation_joint_name(
                target_model=target_model,
                rpy_selection=rpy_selection,
                remove_puppeteer_prefix=remove_puppeteer_prefix,
            )
            for rpy_selection in rpy_equivalent_body_rotation_order()
        ]

    @property
    def translation_axis_names(self) -> List[str]:
        return ["X", "Y", "Z"]

    def translation_joint_names(
        self,
        target_model: ModelInstanceIndex,
        remove_puppeteer_prefix: bool = False,
    ) -> List[str]:
        """
        Description
        -----------
        This method returns the names of all translation joints created for the puppet.
        """
        # Setup
        config = self.config
        plant: MultibodyPlant = self.plant

        # Find name of plant
        target_model_name = plant.GetModelInstanceName(target_model)
        puppeteer_prefix = "" if remove_puppeteer_prefix else f"[{config.name}]"

        return [
            f"{puppeteer_prefix}{target_model_name}_translation_joint_{axis_name}"
            for axis_name in self.translation_axis_names
        ]


from dataclasses import dataclass
from pathlib import Path
from pydrake.geometry.optimization import HPolyhedron
from pydrake.all import ModelInstanceIndex

# Internal Imports
from brom_drake.utils.pick_and_place.phase import PickAndPlacePhase

@dataclass
class PickAndPlaceTargetDescription:
    file_path: Path
    goal_region: HPolyhedron
    model_instance_index: ModelInstanceIndex = None
    _name: str = None

    @property
    def name(self) -> str:
        # If a name was provided, then provide that instead.
        if self._name is not None:
            return self._name
        
        # If a name was not provided, then extract a name from the file_path.
        file_name = self.file_path.name
        return file_name[:file_name.find('.')]
        
    def name_for_pick_and_place_phase(self, phase: PickAndPlacePhase) -> str:
        return self.name + "_" + phase.to_str()



from pathlib import Path
import xml.etree.ElementTree as ET
from typing import Union

URDF_CONVERSION_LOG_LEVEL_NAME = "BROM_URDF_CONVERSION"
URDF_CONVERSION_LEVEL = 21

def does_drake_parser_support(filename: str) -> bool:
    """
    *Description*
    
    This function cleanly answers whether the given 3d object
    file is supported by Drake.
    
    *Parameters*

    filename: str
        The name of the file to check.
    
    *Returns*

    does_support: bool
        True if the file is supported by Drake, False otherwise.
    
    """
    return ".obj" in filename # TODO(kwesi): Determine if .sdf should be put here.

def create_transmission_element_for_joint(
    actuated_joint_name: str,
) -> ET.Element:
    """
    *Description*
    
    This function creates a transmission element for the given joint.

    *Parameters*

    actuated_joint_name: str
        The name of the joint that will be actuated

    *Returns*

    transmission_element: xml.etree.ElementTree.Element
        The transmission element for the given joint.
    """

    # Create the transmission element
    transmission = ET.Element("transmission")

    # Create an inner type element
    transmission_type_element = ET.Element("type")
    transmission_type_element.text = "transmission_interface/SimpleTransmission"

    # Create inner joint element (reference to the actuated_joint_name
    joint_element = ET.Element("joint")
    joint_element.set("name", actuated_joint_name)

    # Create inner actuator element
    actuator_element = ET.Element("actuator")
    actuator_element.set("name", f"{actuated_joint_name}_actuator")

    # Assemble transmission element
    transmission.append(transmission_type_element)
    transmission.append(joint_element)
    transmission.append(actuator_element)

    return transmission

def tree_contains_transmission_for_joint(
    tree: ET.ElementTree,
    actuated_joint_name: str,
) -> bool:
    """
    *Description*
    
    This function determines if the given tree contains a transmission
    element for the given actuated joint.

    *Parameters*

    tree: xml.etree.ElementTree.ElementTree
        The tree to search for the transmission element.

    actuated_joint_name: str
        The name of the joint that will be actuated.

    *Returns*

    transmission_exists: bool
        True if the tree contains a transmission element for the given joint,
    """
    # Setup
    root = tree.getroot()

    # Check to see if the root is a transmission element
    if root.tag == "transmission":
        # Check if the transmission element contains the actuated joint
        for child in root:
            if child.tag == "joint" and child.attrib["name"] == actuated_joint_name:
                return True

    # Check each of the tree's children to see if there is a transmission element
    for child in root:
        child_tree_contains_transmission_for_joint = tree_contains_transmission_for_joint(
            ET.ElementTree(child),
            actuated_joint_name,
        )
        if child_tree_contains_transmission_for_joint:
            return True

    # If we've searched through the full sub-tree and don't see the actuated joint,
    # then we return False
    return False

def get_mesh_element_in(collision_element: ET.Element) -> Union[ET.Element, None]:
    """
    *Description*
    
    This function finds the mesh element in the given collision element.

    *Parameters*
    
    collision_element: ET.Element
        The collision element to search for a mesh element
    
    *Returns*
   
    new_element: ET.Element or None
        The mesh element if found, otherwise None
    """
    # Check if the collision element has a <geometry> child
    geometry = collision_element.find("geometry")
    if geometry is not None:
        # Check if the geometry has a <mesh> child
        mesh = geometry.find("mesh")
        if mesh is not None:
            return mesh
    
    # If no mesh element is found, return None
    return None

def find_mesh_file_path_in(collision_element: ET.Element) -> Union[Path, None]:
    """
    *Description*
    
    This function finds the mesh filename in the given collision element.

    *Parameters*
    
    collision_element: ET.Element
        The collision element to search for a mesh filename
    
    *Returns*
    
    mesh_file_path: Path|None
        The mesh filename if found, otherwise None
    """
    # Setup

    # Algorithm
    mesh_elt = get_mesh_element_in(collision_element)
    if mesh_elt is not None:
        # Return the filename attribute of the mesh element
        return Path(
            mesh_elt.attrib.get("filename", None)
        )
    
    # If no mesh filename is found, return None
    return None



from typing import List, Set, Tuple

class EdgeContainsNoLabelsError(ValueError):
    """
    Description
    -----------
    Raised when an edge in the FSM graph contains no labels.
    """
    def __init__(self, edge: Tuple[int, int]):
        self.edge = edge
        self.message = f"Edge {edge} contains no labels (i.e., it is an empty dictionary);\n"
        self.message += f"Need at least one label (the label should be called \"conditions\")!\n"
        super().__init__(self.message)

class EdgeContainsNoConditionsError(ValueError):
    """
    Description
    -----------
    Raised when an edge in the FSM graph contains no conditions.
    """
    def __init__(self, edge: Tuple[int, int]):
        self.edge = edge
        self.message = f"Edge {edge} contains 0 conditions; need at least one!\n"
        super().__init__(self.message)

class MultipleConnectedComponentsError(ValueError):
    """
    Description
    -----------
    Raised when the FSM graph has multiple connected components.
    """
    def __init__(self, connected_components: List[Set[int]]):
        self.connected_components = connected_components
        self.message = f"Found {len(connected_components)} connected components in the FSM graph.\n"
        self.message += f"Connected components are: {connected_components}.\n"
        super().__init__(self.message)

class NumberOfStartNodesError(ValueError):
    """
    Description
    -----------
    Raised when the FSM graph has multiple start nodes.
    """
    def __init__(self, start_nodes: List[int]):
        self.start_nodes = start_nodes
        self.message = f"Found {len(start_nodes)} start nodes in the FSM graph.\n"
        self.message += f"Start nodes are: {start_nodes}.\n"
        super().__init__(self.message)

class OutputPortNotInitializedError(ValueError):
    """
    Description
    -----------
    Raised when an output port is not intiialized in the FSM graph.
    """
    def __init__(self, node_index: int, output_port: str, output_ports_initialized: List[str]):
        self.output_port = output_port
        self.message = f"Output port \"{output_port}\" for node #{node_index} does not have an initial value in the FSM graph;\n"
        self.message += f"The ports that were initialized are {output_ports_initialized}!\n"
        super().__init__(self.message)


from pydrake.all import AbstractValue, Context, LeafSystem, PiecewisePose, RigidTransform

class PoseTrajectorySource(LeafSystem):
    def __init__(self, trajectory: PiecewisePose):
        LeafSystem.__init__(self)

        # Setup
        self._trajectory = trajectory

        # Create output port
        initial_pose = self._trajectory.GetPose(0.0)
        self.DeclareAbstractOutputPort(
            "current_pose",
            lambda: AbstractValue.Make(initial_pose),
            self.get_current_pose,
        )

    def get_current_pose(self, context: Context, output: AbstractValue):
        # Get the current time
        t = context.get_time()

        # Evaluate the trajectory at the current time
        current_pose = self._trajectory.GetPose(t)

        # Set the output
        output.set_value(current_pose)



from enum import IntEnum

class PickAndPlacePhase(IntEnum):
    kPreGrasp = 1
    kGrasp = 2
    kPostGrasp = 3
    kPrePlace = 4
    kPlace = 5
    kPostPlace = 6

    def to_str(self) -> str:
        match self:
            case PickAndPlacePhase.kPreGrasp:
                return "pre_grasp"
            case PickAndPlacePhase.kGrasp:
                return "grasp"
            case PickAndPlacePhase.kPostGrasp:
                return "post_grasp"
            case PickAndPlacePhase.kPrePlace:
                return "pre_place"
            case PickAndPlacePhase.kPlace:
                return "place"
            case PickAndPlacePhase.kPostPlace:
                return "post_place"
            case _:
                raise NotImplementedError(
                    f"The string for PickAndPlacePhase \"{self}\" does not exist yet. Create an issue on GitHub to see this addressed!"
                )
            